#!/usr/bin/env python
from confindr_src.methods import check_acceptable_xmx, check_for_databases_and_download, check_valid_base_fraction, \
    dependency_check, find_paired_reads, find_unpaired_reads, find_contamination, get_version, write_output
import multiprocessing
import subprocess
import traceback
import argparse
import logging
import shutil
import os


def confindr(args):
    # Check for dependencies.
    all_dependencies_present = True
    # Re-enable minimap2 as dependency once nanopore stuff actually works.
    if args.data_type == 'Illumina':
        dependencies = ['bbmap.sh', 'bbduk.sh', 'mash', 'kma']
    else:
        dependencies = ['bbduk.sh', 'mash', 'minimap2', 'kma']

    for dependency in dependencies:
        if dependency_check(dependency) is False:
            logging.error('Dependency {} not found. Please make sure it is installed and present'
                          ' on your $PATH.'.format(dependency))
            all_dependencies_present = False
    if not all_dependencies_present:
        logging.error('Could not find all necessary dependencies, quitting...')
        quit(code=1)

    # Check that the base fraction specified actually makes sense.
    if check_valid_base_fraction(args.base_fraction_cutoff) is False:
        logging.error('Base fraction must be between 0 and 1 if specified. Input value was: {}'
                      .format(args.base_fraction_cutoff))
        quit(code=1)

    # If user specified Xmx, make sure that they actually entered a value that will work. If not, the method will tell
    # them what they did wrong. Then quit.
    if args.Xmx:
        valid_xmx = check_acceptable_xmx(args.Xmx)
        if valid_xmx is False:
            quit(code=1)

    # Don't yet have cgmlst support with Nanopore reads - don't let user do this.
    if args.cgmlst and args.data_type == 'Nanopore':
        logging.error('ERROR: cgMLST schemes not yet supported for Nanopore reads. Quitting...')
        quit(code=1)

    if args.data_type == 'Nanopore':
        logging.warning('WARNING: Nanopore contamination detection is highly experimental. Any results should be taken '
                        'with several very large grains of salt. If you are going to try this, try setting -q to '
                        'somewhere in the range of 12-15, and only count things as contaminated that have at least '
                        '10 contaminating SNVs. Even then, results may be wonky. In particular, samples with lots of '
                        'depth will probably always show up as contaminated.')

    # Make the output directory.
    if not os.path.isdir(args.output_name):
        os.makedirs(args.output_name)
    # Remove any reports created by previous iterations of ConFindr
    try:
        os.remove(os.path.join(args.output_name, 'confindr_report.csv'))
    except FileNotFoundError:
        pass
    # Set the minimum number of matching hashes
    min_matching_hashes = args.min_matching_hashes
    # Check if databases necessary to run are present, and download them if they aren't
    check_for_databases_and_download(database_location=args.databases)

    # Figure out what pairs of reads, as well as unpaired reads, are present.
    paired_reads = find_paired_reads(args.input_directory,
                                     forward_id=args.forward_id,
                                     reverse_id=args.reverse_id)
    unpaired_reads = find_unpaired_reads(args.input_directory,
                                         forward_id=args.forward_id,
                                         reverse_id=args.reverse_id,
                                         find_fasta=args.fasta)
    # Consolidate read lists
    reads = sorted(paired_reads + unpaired_reads)
    # Process paired reads, one sample at a time.
    for fastq in reads:
        if len(fastq) == 1:
            sample_name = os.path.split(fastq[0])[-1].split('.')[0]
        else:
            sample_name = os.path.split(fastq[0])[-1].split(args.forward_id)[0]
        logging.info('Beginning analysis of sample {}...'.format(sample_name))
        try:
            find_contamination(pair=fastq,
                               forward_id=args.forward_id,
                               threads=args.threads,
                               output_folder=args.output_name,
                               databases_folder=args.databases,
                               keep_files=args.keep_files,
                               quality_cutoff=args.quality_cutoff,
                               base_cutoff=args.base_cutoff,
                               base_fraction_cutoff=args.base_fraction_cutoff,
                               cgmlst_db=args.cgmlst,
                               xmx=args.Xmx,
                               tmpdir=args.tmp,
                               data_type=args.data_type,
                               use_rmlst=args.rmlst,
                               min_matching_hashes=min_matching_hashes,
                               fasta=args.fasta,
                               debug=args.verbosity)
        except subprocess.CalledProcessError:
            # If something unforeseen goes wrong, traceback will be printed to screen.
            # We then add the sample to the report with a note that it failed.
            multi_positions = 0
            genus = 'Error processing sample'
            write_output(output_report=os.path.join(args.output_name, 'confindr_report.csv'),
                         sample_name=sample_name,
                         multi_positions=multi_positions,
                         genus=genus,
                         total_gene_length=0,
                         database_download_date='ND')
            logging.warning('Encountered error when attempting to run ConFindr on sample '
                            '{sample}. Skipping...'.format(sample=sample_name))
            logging.warning('Error encountered was:\n{}'.format(traceback.format_exc()))
            if args.keep_files is False:
                shutil.rmtree(os.path.join(args.output_name, sample_name))
    if args.keep_files is False and args.tmp is not None:
        shutil.rmtree(args.tmp)
    logging.info('Contamination detection complete!')


def main():
    version = get_version()
    cpu_count = multiprocessing.cpu_count()
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input_directory',
                        type=str,
                        required=True,
                        help='Folder that contains fastq files you want to check for contamination. '
                             'Will find any file that contains .fq or .fastq in the filename.')
    parser.add_argument('-o', '--output_name',
                        type=str,
                        required=True,
                        help='Base name for output/temporary directories.')
    parser.add_argument('-d', '--databases',
                        type=str,
                        default=os.environ.get('CONFINDR_DB', os.path.expanduser('~/.confindr_db')),
                        help='Databases folder. To download these, you will need to get access to the rMLST databases. '
                             'For complete instructions on how to do this, please see '
                             'https://olc-bioinformatics.github.io/ConFindr/install/#downloading-confindr-databases')
    parser.add_argument('--rmlst',
                        default=False,
                        action='store_true',
                        help='Activate to prefer using rMLST databases over core-gene derived databases. By default,'
                             'ConFindr will use core-gene derived databases where available.')
    parser.add_argument('-t', '--threads',
                        type=int,
                        default=cpu_count,
                        help='Number of threads to run analysis with.')
    parser.add_argument('-tmp', '--tmp',
                        type=str,
                        help='If your ConFindr databases are in a location you don\'t have write access to, '
                             'you can enter this option to specify a temporary directory to put genus-specific '
                             'databases to.')
    parser.add_argument('-k', '--keep_files',
                        default=False,
                        action='store_true',
                        help='By default, intermediate files are deleted. Activate this flag to keep intermediate '
                             'files.')
    parser.add_argument('-q', '--quality_cutoff',
                        type=int,
                        default=20,
                        help='Base quality needed to support a multiple allele call. Defaults to 20.')
    parser.add_argument('-b', '--base_cutoff',
                        type=int,
                        default=3,
                        help='Number of bases necessary to support a multiple allele call, and automatically '
                        'increments based upon gene-specific quality score, length and depth of coverage. '
                        'Default is 3.')
    parser.add_argument('-bf', '--base_fraction_cutoff',
                        type=float,
                        default=0.05,
                        help='Fraction of bases necessary to support a multiple allele call. Particularly useful when '
                             'dealing with very high coverage samples. Default is 0.05.')
    parser.add_argument('-e', '--error_cutoff',
                        type=float,
                        default=1.0,
                        help='Value to use for the calculated error cutoff when setting the base cutoff value. '
                             'Default is 1.0%%.')
    parser.add_argument('-fid', '--forward_id',
                        type=str,
                        default='_R1',
                        help='Identifier for forward reads.')
    parser.add_argument('-rid', '--reverse_id',
                        type=str,
                        default='_R2',
                        help='Identifier for reverse reads.')
    parser.add_argument('-v', '--version',
                        action='version',
                        version=version)
    parser.add_argument('-dt', '--data_type',
                        choices=['Illumina', 'Nanopore'],
                        default='Illumina',
                        help='Type of input data. Default is Illumina, but can be used for Nanopore too. No PacBio '
                             'support (yet).')
    parser.add_argument('-Xmx', '--Xmx',
                        type=str,
                        help='Very occasionally, parts of the pipeline that use the BBMap suite will have their memory '
                             'reservation fail and request not enough, or sometimes negative, memory. If this happens '
                             'to you, you can use this flag to override automatic memory reservation and use an amount '
                             'of memory requested by you. -Xmx 20g will specify 20 gigs of RAM, and -Xmx 800m '
                             'will specify 800 megs.')
    parser.add_argument('-cgmlst', '--cgmlst',
                        type=str,
                        help='Path to a cgMLST database to use for contamination detection instead of using the default'
                             ' rMLST database. Sequences in this file should have headers in format '
                             '>genename_allelenumber. To speed up ConFindr runs, clustering the cgMLST database with '
                             'CD-HIT before running ConFindr is recommended. This is highly experimental, results '
                             'should be interpreted with great care.')
    parser.add_argument('--fasta',
                        default=False,
                        action='store_true',
                        help='If activated, will look for FASTA files instead of FASTQ for unpaired reads.')
    parser.add_argument('-verbosity', '--verbosity',
                        choices=['debug', 'info', 'warning'],
                        default='info',
                        help='Amount of output you want printed to the screen. Defaults to info, which should be good '
                             'for most users.')
    parser.add_argument('-m', '--min_matching_hashes',
                        default=150,
                        type=int,
                        help='Minimum number of matching hashes in a MASH screen in order for a genus to be considered '
                             'present in a sample. Default is 150')
    args = parser.parse_args()
    # Setup the logger. TODO: Different colors for different levels.
    if args.verbosity == 'info':
        logging.basicConfig(format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
                            level=logging.INFO,
                            datefmt='%Y-%m-%d %H:%M:%S')
    elif args.verbosity == 'debug':
        logging.basicConfig(format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
                            level=logging.DEBUG,
                            datefmt='%Y-%m-%d %H:%M:%S')
    elif args.verbosity == 'warning':
        logging.basicConfig(format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
                            level=logging.WARNING,
                            datefmt='%Y-%m-%d %H:%M:%S')

    logging.info('Welcome to {version}! Beginning analysis of your samples...'.format(version=version))
    confindr(args)


if __name__ == '__main__':
    main()



from confindr_src.methods import *
import subprocess
import pytest
import shutil
import csv
import os

parentdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
os.sys.path.insert(0, parentdir)

#---------------------------
# Test ConFindr integration
#---------------------------

def test_integration():
    correct_contamination_calls = {'SRX5084910_SRR8268082': 'True',
                                   'SRX5084911_SRR8268081': 'False',
                                   'SRX5084914_SRR8268078': 'True',
                                   'SRX5084915_SRR8268077': 'False',
                                   'SRX5084941_SRR8268051': 'True',
                                   'SRX5084940_SRR8268052': 'False',
                                   'SRX5084995_SRR8267997': 'True'
                                   }
    correct_genera = {'SRX5084910_SRR8268082': 'Escherichia',
                      'SRX5084911_SRR8268081': 'Escherichia',
                      'SRX5084914_SRR8268078': 'Salmonella',
                      'SRX5084915_SRR8268077': 'Salmonella',
                      'SRX5084940_SRR8268052': 'Listeria',
                      'SRX5084941_SRR8268051': 'Listeria',
                      'SRX5084995_SRR8267997': 'Salmonella:Citrobacter'}
    subprocess.call("confindr.py -i tests/test_samples -o confindr_integration_output -d databases -k -fid '_1' -rid '_2'",
                     shell=True)
    with open('confindr_integration_output/confindr_report.csv') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            sample = row['Sample']
            if 'cross_contaminated' not in sample:
                assert row['ContamStatus'] == correct_contamination_calls[sample]
                assert row['Genus'] == correct_genera[sample]
            else:
                assert row['ContamStatus'] == correct_contamination_calls[sample]
                genera = row['Genus'].split(':')
                assert 'Salmonella' in genera and 'Escherichia' in genera and 'Listeria' in genera
    shutil.rmtree('confindr_integration_output')
    shutil.rmtree('databases')

#-------------------
# Test dependencies
#-------------------

def test_present_dependency():
    assert dependency_check('ls') is True


def test_nonexistent_dependency():
    assert dependency_check('fake_dependency') is False

#-------------------------------
# Test file pattern recognition
#-------------------------------

def test_r1_fastqs():
    assert find_paired_reads('tests/fake_fastqs/') == [['tests/fake_fastqs/test_R1.fastq.gz',
                                                        'tests/fake_fastqs/test_R2.fastq.gz']]

def test_1_fastqs():
    assert find_paired_reads('tests/fake_fastqs/', forward_id='_1',
                             reverse_id='_2') == [['tests/fake_fastqs/test_1.fastq.gz',
                                                   'tests/fake_fastqs/test_2.fastq.gz']]

def test_empty_fastqs():
    assert find_paired_reads('tests/fake_fastqs/', forward_id='_asdf', reverse_id='_fdsa') == []


def test_unpaired_fastq():
    assert ['tests/fake_fastqs/test_alone.fastq.gz'] == find_unpaired_reads('tests/fake_fastqs')[2]

#-------------------------
# Test subprocess calling
#-------------------------

def test_run_cmd_success():
    cmd = 'echo asdf'
    out, err = run_cmd(cmd)
    assert out == 'asdf\n'
    assert err == ''

def test_run_cmd_failure_exit_code():
    with pytest.raises(subprocess.CalledProcessError):
        run_cmd('garbagecommandthatdoesnotwork')

#----------------------------------
# Test base count cutoff reporting
#----------------------------------

def test_two_hq_bases_above_threshold():
    assert number_of_bases_above_threshold({'G': 80, 'A': 20}) == 2


def test_just_one_hq_bases_above_threshold():
    assert number_of_bases_above_threshold({'G': 99, 'A': 1}) == 1


def test_two_hq_bases_above_threshold_custom_params():
    assert number_of_bases_above_threshold({'G': 99, 'A': 1}, base_count_cutoff=1) == 2


def test_just_one_hq_base_above_threshold_custom_params():
    assert number_of_bases_above_threshold({'G': 96, 'A': 4}, base_count_cutoff=5) == 1


def test_three_hq_bases_above_threshold():
    assert number_of_bases_above_threshold({'G': 90, 'A': 10, 'T': 10}) == 3


def test_two_out_of_three_hq_bases_above_threshold():
    assert number_of_bases_above_threshold({'G': 90, 'A': 9, 'T': 1}) == 2

#-------------------------------------
# Test base fraction cutoff reporting
#-------------------------------------

def test_two_hq_bases_above_fraction_threshold():
    assert number_of_bases_above_threshold({'G': 80, 'A': 20}, base_fraction_cutoff=0.05) == 2


def test_two_hq_bases_above_fraction_threshold_low_coverage():
    assert number_of_bases_above_threshold({'G': 9, 'A': 1}, base_fraction_cutoff=0.05) == 1


def test_two_hq_bases_above_fraction_threshold_low_coverage_one_base_counts():
    assert number_of_bases_above_threshold({'G': 9, 'A': 1}, base_count_cutoff=1, base_fraction_cutoff=0.05) == 2


def test_just_one_hq_bases_above_fraction_threshold():
    assert number_of_bases_above_threshold({'G': 99, 'A': 1}, base_fraction_cutoff=0.05) == 1


def test_three_hq_bases_above_fraction_threshold():
    assert number_of_bases_above_threshold({'G': 90, 'A': 10, 'T': 10}, base_fraction_cutoff=0.05) == 3


def test_two_out_of_three_hq_bases_above_fraction_threshold():
    assert number_of_bases_above_threshold({'G': 90, 'A': 9, 'T': 1}, base_fraction_cutoff=0.05) == 2


def test_valid_base_fraction_none():
    assert check_valid_base_fraction(None) is True


def test_valid_base_fraction_zero():
    assert check_valid_base_fraction(0.0) is True


def test_valid_base_fraction_one():
    assert check_valid_base_fraction(1.0) is True


def test_valid_base_fraction_between_zero_one():
    assert check_valid_base_fraction(0.2) is True


def test_invalid_base_fraction():
    assert check_valid_base_fraction(1.2) is False

#----------------------------------------
# Test total sequence length calculation
#----------------------------------------

def test_total_length_fasta():
    assert find_total_sequence_length('tests/rmlst.fasta') == 20862

#----------------------
# Test report creation
#----------------------

def test_write_output_creates_file_if_does_not_exist():
    write_output(output_report='tests/confindr_report.csv',
                 sample_name='Test',
                 multi_positions=55,
                 genus='Fakella',
                 total_gene_length=888,
                 database_download_date='ND')
    assert os.path.isfile('tests/confindr_report.csv') is True
    # os.remove('tests/confindr_report.csv')

def test_write_output_appends_if_file_does_exist():
    write_output(output_report='tests/confindr_report.csv',
                 sample_name='Test',
                 multi_positions=55,
                 genus='Fakella',
                 total_gene_length=888,
                 database_download_date='ND')
    with open('tests/confindr_report.csv') as f:
        lines = f.readlines()
    assert len(lines) > 2

#----------------------------------------------
# Test converting base dictionaries to strings
#----------------------------------------------

def test_base_dict_to_string_two_base_descending():
    assert base_dict_to_string({'A': 18, 'C': 3}) == 'A:18;C:3'


def test_base_dict_to_string_two_base_ascending():
    assert base_dict_to_string({'A': 8, 'C': 33}) == 'C:33;A:8'


def test_base_dict_to_string_three_bases():
    assert base_dict_to_string({'A': 5, 'T': 88, 'C': 33}) == 'T:88;C:33;A:5'

#----------------------------------------------
# Check that BBTools memory settings are valid
#----------------------------------------------

def test_valid_xmx_string_gigabytes():
    assert check_acceptable_xmx('20g') is True
    assert check_acceptable_xmx('20G') is True


def test_valid_xmx_string_megabytes():
    assert check_acceptable_xmx('20m') is True
    assert check_acceptable_xmx('20M') is True


def test_valid_xmx_string_kilobytes():
    assert check_acceptable_xmx('550k') is True
    assert check_acceptable_xmx('550K') is True


def test_invalid_xmx_bad_suffix():
    assert check_acceptable_xmx('600u') is False


def test_invalid_xmx_float():
    assert check_acceptable_xmx('2.2G') is False


def test_invalid_xmx_not_an_integer():
    assert check_acceptable_xmx('asdfK') is False



#!/usr/bin/env python
from subprocess import Popen, PIPE


def run_subprocess(command):
    """
    command is the command to run, as a string.
    runs a subprocess, returns stdout and stderr from the subprocess as strings.
    """
    x = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
    out, err = x.communicate()
    out = out.decode('utf-8')
    err = err.decode('utf-8')
    return out, err


class MashResult:
    def __init__(self, mash_result_row):
        x = mash_result_row.split()
        self.reference = x[0]
        self.query = x[1]
        self.distance = float(x[2])
        self.pvalue = float(x[3])
        self.matching_hash = x[4]


class ScreenResult:
    def __init__(self, screen_result_row):
        x = screen_result_row.split()
        self.identity = float(x[0])
        self.shared_hashes = x[1]
        self.median_multiplicity = x[2]
        self.pvalue = float(x[3])
        self.query_id = x[4]


def kwargs_to_string(kwargs):
    """
    Given a set of kwargs, turns them into a string which can then be passed to a command.
    :param kwargs: kwargs from a function call.
    :return: outstr: A string, which is '' if no kwargs were given, and the kwargs in string format otherwise.
    """
    outstr = ''
    for arg in kwargs:
        outstr += ' -{} {}'.format(arg, kwargs[arg])
    return outstr


def sketch(*args, output_sketch='sketch.msh', threads=1, returncmd=False, **kwargs):
    """
    Wrapper for mash sketch.
    :param args: Files you want to sketch. Any number can be passed in, file patterns (i.e. *fasta) can be used.
    :param output_sketch: Output file for your sketch. Default sketch.msh.
    :param threads: Number of threads to run analysis on.
    :param kwargs: Other arguments, in parameter='argument' format. If parameter is just a switch, do parameter=''
    :param returncmd: If true, will return the command used to call mash as well as out and err.
    :return: stdout and stderr from mash sketch
    """
    options = kwargs_to_string(kwargs)
    if len(args) == 0:
        raise ValueError('At least one file to sketch must be specified. You specified 0 files.')
    cmd = 'mash sketch '
    for arg in args:
        cmd += arg + ' '
    cmd += '-o {} -p {} {}'.format(output_sketch, str(threads), options)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def dist(*args, output_file='distances.tab', threads=1, returncmd=False, **kwargs):
    """
    Wrapper for mash dist.
    :param args: Files you want to find distances between. Can be
    :param output_file: Output file to write your distances to. Default distances.tab
    :param threads: Number of threads to run mash on.
    :param kwargs: Other arguments, in parameter='argument' format. If parameter is just a switch, do parameter=''
    :param returncmd: If true, will return the command used to call mash as well as out and err.
    :return: stdout and stderr from mash dist
    """
    options = kwargs_to_string(kwargs)
    if len(args) == 0:
        raise ValueError('At least one file to sketch must be specified. You specified 0 files.')
    cmd = 'mash dist '
    for arg in args:
        cmd += arg + ' '
    cmd += ' -p {} {} > {}'.format(str(threads), options, output_file)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def screen(*args, output_file='screen.tab', threads=1, returncmd=False, **kwargs):
    """
    Wrapper for mash screen. Requires mash v2.0 or higher.
    :param args: Files you want to screen. First argument must be a sketch.
    :param output_file: Output to write containment info to.
    :param threads: Number of threads to run mash on.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param kwargs: Other arguments, in parameter='argument' format. If parameter is just a switch, do parameter=''
    :return: stdout and stderr from mash screen
    """
    options = kwargs_to_string(kwargs)
    cmd = 'mash screen '
    for arg in args:
        cmd += arg + ' '
    cmd += ' -p {} {} | sort -gr > {}'.format(str(threads), options, output_file)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def read_mash_output(result_file):
    """
    :param result_file: Tab-delimited result file generated by mash dist.
    :return: mash_results: A list with each entry in the result file as an entry, with attributes reference, query,
    distance, pvalue, and matching_hash
    """
    with open(result_file) as handle:
        lines = handle.readlines()
    mash_results = list()
    for line in lines:
        result = MashResult(line)
        mash_results.append(result)
    return mash_results


def read_mash_screen(screen_result):
    """
    :param screen_result: Tab-delimited result file generated by mash screen.
    :return: results: A list with each line in the result file as an entry, with attributes identity, shared_hashes,
    median_multiplicity, pvalue, and query_id
    """
    with open(screen_result) as handle:
        lines = handle.readlines()
    results = list()
    for line in lines:
        result = ScreenResult(line)
        results.append(result)
    return results



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name="confindr",
    version="0.8.1",
    packages=find_packages(),
    entry_points={
       'console_scripts': [
            'confindr.py = confindr_src.confindr:main',
            'confindr = confindr_src.confindr:main',
            'confindr_database_setup = confindr_src.database_setup:main',
            'confindr_create_db = confindr_src.create_genus_specific_db:main'
       ],
    },
    author="Adam Koziol",
    author_email="adam.koziol@inspection.gc.ca",
    url="https://github.com/OLC-Bioinformatics/ConFindr",
    install_requires=['biopython',
                      'pysam',
                      'pytest',
                      'numpy',
                      'rauth']
)



#!/usr/bin/env python
from confindr_src.wrappers import bbtools, mash
from Bio import SeqIO
from pysam.utils import SamtoolsError
from itertools import chain
from statistics import mean
import multiprocessing
import urllib.request
import pkg_resources
import numpy as np
import subprocess
import logging
import shutil
import tarfile
import pysam
import glob
import gzip
import math
import csv
import os


def download_mash_sketch(output_folder):
    logging.info('Downloading mash refseq sketch...')
    urllib.request.urlretrieve('https://github.com/OLC-Bioinformatics/ConFindr/raw/master/refseq_sketch/refseq.msh',
                               os.path.join(output_folder, 'refseq.msh'))


def download_cgmlst_derived_data(output_folder):
    logging.info('Downloading cgMLST-derived data for Escherichia, Salmonella, and Listeria...')
    urllib.request.urlretrieve('https://ndownloader.figshare.com/files/14771267',
                               os.path.join(output_folder, 'confindr_db.tar.gz'))
    confindr_tar = os.path.join(output_folder, 'confindr_db.tar.gz')
    tar = tarfile.open(confindr_tar)
    tar.extractall(path=output_folder)
    tar.close()
    os.remove(confindr_tar)
    index(output_folder=output_folder,
          genera=['Escherichia', 'Listeria', 'Salmonella'],
          cgderived=True)


def index(output_folder, genera, cgderived=False):
    """

    :param output_folder:
    :param genera:
    :param cgderived:
    :return:
    """
    for predominant_genus in genera:
        if cgderived:
            sample_database = os.path.join(output_folder, '{pg}_db_cgderived.fasta'.format(pg=predominant_genus))
        else:
            sample_database = os.path.join(output_folder, '{pg}_db.fasta'.format(pg=predominant_genus))

        if not os.path.isfile(sample_database):

            if os.path.isfile(os.path.join(output_folder, 'gene_allele.txt')) and \
                    os.path.isfile(os.path.join(output_folder, 'rMLST_combined.fasta')):
                logging.info('Setting up rMLST genus-specific database for genus {pg}...'
                             .format(pg=predominant_genus))
                allele_list = find_genus_specific_allele_list(os.path.join(output_folder, 'gene_allele.txt'),
                                                              predominant_genus)
                # Create the allele-specific database
                setup_allelespecific_database(fasta_file=sample_database,
                                              database_folder=output_folder,
                                              allele_list=allele_list)
        # Perform the necessary samtools and KMA indexing
        index_databases(sample_database=sample_database)


def run_cmd(cmd):
    """
    Runs a command using subprocess, and returns both the stdout and stderr from that command
    If exit code from command is non-zero, raises subprocess.CalledProcessError
    :param cmd: command to run as a string, as it would be called on the command line
    :return: out, err: Strings that are the stdout and stderr from the command called.
    """
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    out = out.decode('utf-8')
    err = err.decode('utf-8')
    if p.returncode != 0:
        raise subprocess.CalledProcessError(p.returncode, cmd=cmd)
    return out, err


def write_to_logfile(logfile, out, err, cmd):
    """
    Writes stdout, stderr, and a command to a logfile
    :param logfile: Path to file to write output to.
    :param out: Stdout of program called, as a string
    :param err: Stderr of program called, as a string
    :param cmd: command that was used
    """
    with open(logfile, 'a+') as outfile:
        outfile.write('Command used: {}\n\n'.format(cmd))
        outfile.write('STDOUT: {}\n\n'.format(out))
        outfile.write('STDERR: {}\n\n'.format(err))


def dependency_check(dependency):
    """
    Uses shutil to check if a dependency is installed (won't check version of anything - just presence)
    :param dependency: The dependency as it would be called on the command line (i.e. for blastn, would be blastn)
    :return: True if dependency is present, False if it is not found.
    """
    if shutil.which(dependency) is not None:
        return True
    else:
        return False


def find_paired_reads(fastq_directory, forward_id='_R1', reverse_id='_R2'):
    """
    Looks at a directory to try to find paired fastq files. Should be able to find anything fastq.
    :param fastq_directory: Complete path to directory containing fastq files.
    :param forward_id: Identifier for forward reads. Default R1.
    :param reverse_id: Identifier for reverse reads. Default R2.
    :return: List containing pairs of fastq files, in format [[forward_1, reverse_1], [forward_2, reverse_2]], etc.
    """
    pair_list = list()

    fastq_files = glob.glob(os.path.join(fastq_directory, '*.f*q*'))
    for name in sorted(fastq_files):
        if forward_id in name and os.path.isfile(name.replace(forward_id, reverse_id)):
            pair_list.append([name, name.replace(forward_id, reverse_id)])
    return pair_list


def find_unpaired_reads(fastq_directory, forward_id='_R1', reverse_id='_R2', find_fasta=False):
    """
    Looks at a directory to find unpaired fastq files.
    :param fastq_directory: Complete path to directory containing fastq files.
    :param forward_id: Identifier for forward reads. Default _R1.
    :param reverse_id: Identifier for forward reads. Default _R2.
    :param find_fasta: If False, will look for fastq files. Otherwise, looks for Fasta files.
    :return: List of files that appear to be unpaired reads.
    """
    read_list = list()
    if find_fasta is False:
        fastq_files = glob.glob(os.path.join(fastq_directory, '*.f*q*'))
    else:
        # Very misnamed!
        fastq_files = glob.glob(os.path.join(fastq_directory, '*.f*a*'))
    for name in sorted(fastq_files):
        # Iterate through files, adding them to our list of unpaired reads if:
        # 1) They don't have the forward identifier or the reverse identifier in their name.
        # 2) They have forward but the reverse isn't there.
        # 3) They have reverse but the forward isn't there.
        if forward_id not in name and reverse_id not in name:
            read_list.append([name])
        elif forward_id in name and not os.path.isfile(name.replace(forward_id, reverse_id)):
            read_list.append([name])
        elif reverse_id in name and not os.path.isfile(name.replace(reverse_id, forward_id)):
            read_list.append([name])
    return read_list


def find_genus_specific_allele_list(profiles_file, target_genus):
    """
    A new way of making our specific databases: Make our profiles file have lists of every gene/allele present for
    each genus instead of just excluding a few genes for each. This way, should have much smaller databases
    while managing to make ConFindr a decent bit faster (maybe)
    :param profiles_file: Path to profiles file.
    :param target_genus: Genus you want to make a custom database for (STR)
    :return: List of gene/allele combinations that should be part of species-specific database.
    """
    alleles = list()
    with open(profiles_file) as f:
        lines = f.readlines()
    for line in lines:
        line = line.rstrip()
        genus = line.split(':')[0]
        if genus == target_genus:
            alleles = line.split(':')[1].split(',')[:-1]
    return alleles


def setup_allelespecific_database(fasta_file, database_folder, allele_list):
    """
    Since some genera have some rMLST genes missing, or two copies of some genes, genus-specific databases are needed.
    This will take only the alleles known to be part of each genus and write them to a genus-specific file.
    :param database_folder: Path to folder where rMLST_combined is stored.
    :param fasta_file: Path to fasta file to write allele-specific database to.
    :param allele_list: allele list generated by find_genus_specific_allele_list
    """
    rmlst_index = SeqIO.index(os.path.join(database_folder, 'rMLST_combined.fasta'), 'fasta')
    seqs = list()
    for s in allele_list:
        try:
            seqs.append(rmlst_index[s])
        except KeyError:
            logging.warning('Tried to add {} to allele-specific database, but could not find it.'.format(s))
    try:
        SeqIO.write(seqs, fasta_file, 'fasta')
    except FileNotFoundError:
        pass


def find_cross_contamination(databases, reads, sample_name, tmpdir='tmp', log='log.txt', threads=1,
                             min_matching_hashes=40):
    """
    Uses mash to find out whether or not a sample has more than one genus present, indicating cross-contamination.
    :param databases: A databases folder, which must contain refseq.msh, a mash sketch that has one representative
    per genus from refseq.
    :param reads: Relative path(s) to either unpaired (type STR) or paired (type LIST) FASTQ reads
    :param sample_name:
    :param tmpdir: Temporary directory to store mash result files in.
    :param log: Logfile to write to.
    :param threads: Number of threads to run mash with.
    :param min_matching_hashes: Minimum number of matching hashes in a MASH screen in order for a genus to be
    considered present in a sample. Default is 40
    :return: cross_contam: a bool that is True if more than one genus is found, and False otherwise.
    :return: genera_present: A string. If only one genus is found, string is NA. If more than one genus is found,
    the string is a list of genera present, separated by colons (i.e. for Escherichia and Salmonella found, string would
    be 'Escherichia:Salmonella'
    """
    genera_present = list()
    # Only run the MASH analyses if the screen.tab output file does not already exist
    screen_file = os.path.join(tmpdir, '{sn}_screen.tab'.format(sn=sample_name))
    if not os.path.isfile(screen_file):
        if type(reads) is str:
            out, err, cmd = mash.screen('{database}/refseq.msh'.format(database=databases), reads,
                                        threads=threads,
                                        w='',
                                        i='0.85',
                                        output_file=screen_file,
                                        returncmd=True)
        else:
            out, err, cmd = mash.screen('{database}/refseq.msh'.format(database=databases), reads[0],
                                        reads[1],
                                        threads=threads,
                                        w='',
                                        i='0.85',
                                        output_file=screen_file,
                                        returncmd=True)
        write_to_logfile(log, out, err, cmd)
    screen_output = mash.read_mash_screen(os.path.join(tmpdir, '{sn}_screen.tab'.format(sn=sample_name)))
    for item in screen_output:
        mash_genus = item.query_id.split('/')[-3]
        if 'Shigella' in mash_genus:
            mash_genus = 'Escherichia'
        matching_hashes = int(item.shared_hashes.split('/')[0])
        # Only add the genus to the genera_present list of the number of matching hashes exceeds the cutoff
        if matching_hashes >= min_matching_hashes:
            if mash_genus not in genera_present:
                genera_present.append(mash_genus)
    if len(genera_present) == 1:
        genera_present = genera_present[0]
    elif len(genera_present) == 0:
        genera_present = 'ND'
    else:
        tmpstr = ''
        for mash_genus in genera_present:
            tmpstr += mash_genus + ':'
        genera_present = tmpstr[:-1]
    return genera_present


def number_of_bases_above_threshold(high_quality_base_count, base_count_cutoff=2, base_fraction_cutoff=None):
    """
    Finds if a site has at least two bases of high quality, enough that it can be considered
    fairly safe to say that base is actually there.
    :param high_quality_base_count: Dictionary of count of HQ bases at a position where key is base and values is the
    count of that base.
    :param base_count_cutoff: Number of bases needed to support multiple allele presence.
    :param base_fraction_cutoff: Fraction of bases needed to support multiple allele presence.
    :return: True if site has at least base_count_cutoff/base_fraction_cutoff bases, False otherwise
    (changeable by user)
    """

    # make a dict by dictionary comprehension where values are True or False for each base depending on whether the
    # count meets the threshold.
    # Method differs depending on whether absolute or fraction cutoff is specified
    if base_fraction_cutoff:
        total_hq_base_count = sum(high_quality_base_count.values())
        bases_above_threshold = {base: float(count) / total_hq_base_count >= base_fraction_cutoff and count
                                 >= base_count_cutoff for (base, count) in high_quality_base_count.items()}
    else:
        bases_above_threshold = {base: count >= base_count_cutoff for (base, count) in high_quality_base_count.items()}

    # True is equal to 1 so sum of the number of Trues in the bases_above_threshold dict is the number of bases
    # passing threshold
    return sum(bases_above_threshold.values())


def parse_bam(bamfile_name, contig_name, pysam_fasta):
    """
    Use pysam to load the sorted BAM-formatted file, and subsequently the gene-specific read pileup
    :param bamfile_name: Name and path of the sorted BAM file
    :param contig_name: Name of the current gene
    :param pysam_fasta: pysam.FastaFile object created from the reference gene sequence
    :return: bamfile: pysam.AlignmentFile object of the BAM-formatted file
    :return pileup: A pysam pileup object created from the BAM file
    """
    # Load the sorted BAM-formatted file using pysam
    bamfile = pysam.AlignmentFile(bamfile_name, 'rb')
    # These parameters seem to be fairly undocumented with pysam, but I think that they should make the output
    # that I'm getting to match up with what I'm seeing in Tablet.
    pileup = bamfile.pileup(contig_name,
                            stepper='samtools',
                            ignore_orphans=False,
                            fastafile=pysam_fasta,
                            min_base_quality=0)
    return bamfile, pileup


def characterise_read(column, reference_sequence, fastq_records, quality_cutoff,
fasta=False, nanopore=False):
    """
    Parses a column to characterize all the bases present. Determines the number of bases that fit certain criteria
    :param column: A pileupColumn generated by pysam
    :param reference_sequence: String of the FASTA reference gene sequence
    :param fastq_records: Dictionary of SeqIO records parsed from filtered FASTQ reads
    :param quality_cutoff: Desired min phred quality for a base in order to be counted towards a multi-allelic column
    If specified, both the base_cutoff and base_fraction_cutoff will have to be met
    :param fasta: Boolean of whether the samples are in FASTQ or FASTA format
    :param nanopore: Boolean of whether Nanopore reads were provided. Default is False.
    :return: filtered_read_dict: Dictionary with of base types: read direction: count
    :return: qualities: List of all phred scores in the pileup
    """
    # Initialise a dictionary to store the base details
    filtered_read_dict = {
        'congruent_SNV': dict(),
        'congruent_ref': dict(),
        'forward_SNV_reverse_SNV1': dict(),
        'reverse_SNV_forward_SNV1': dict(),
        'forward_SNV_reverse_ref': dict(),
        'reverse_SNV_forward_ref': dict(),
        'forward_SNV_reverse_UM_QF': dict(),
        'forward_ref_reverse_UM_QF': dict(),
        'forward_quality_filtered': dict(),
        'reverse_SNV_forward_UM_QF': dict(),
        'reverse_ref_forward_UM_QF': dict(),
        'reverse_quality_filtered': dict()
    }
    # Initialise a dictionary to store the details parsed from the pileup
    unfiltered_read_details = dict()
    # Initialise a list to store all the phred scores for the bases passing filter in the column
    qualities = list()
    # Iterate through every read present in the column of the pileup
    for read in column.pileups:
        # Not entirely sure why this is sometimes None, but it causes bad stuff
        if read.query_position is not None:
            #  Initialise the read name in the dictionary as required
            if read.alignment.qname not in unfiltered_read_details:
                unfiltered_read_details[read.alignment.qname] = dict()
            # Extract the sequence of the base in the read
            query_base = read.alignment.query_sequence[read.query_position]
            # Extract the sequence of the base in the reference gene
            ref_base = reference_sequence[column.pos]
            # Create a boolean of whether the query base matches the reference base (is it a SNV?)
            match = query_base == ref_base
            # Read names in the pileup have the direction removed - add it back for future parsing. Not an issue for
            # FASTA files
            if not fasta and not nanopore:
                read_name = read.alignment.qname.split(' ')[0] + '/1' if read.alignment.is_read1 else \
                    read.alignment.qname.split(' ')[0] + '/2'
            else:
                read_name = read.alignment.qname
            # Extract the phred quality score from the FASTQ records
            quality = fastq_records[read_name].letter_annotations["phred_quality"][read.query_position]
            # Initialise a dictionary to store the ran
            range_dict = dict()
            # Determine whether there are SNVs clustered together - they will be discarded from the analysis
            # Iterate through a range of the five positions preceding, and five subsequent
            # positions
            for iterator, contig_pos in enumerate(
                    chain(range(column.pos - 5, column.pos), range(column.pos + 1, column.pos + 6))):
                # Ensure that the contig position being examined isn't beyond the length of the gene
                if 0 <= contig_pos < len(reference_sequence) - 1:
                    # Calculate the read position corresponding to the current column position
                    read_pos = list(chain(range(read.query_position - 5, read.query_position),
                                          range(read.query_position + 1, read.query_position + 6)))[iterator]
                    # Ensure that read position being examined isn't beyond the length of the read
                    if 0 <= read_pos < read.alignment.query_alignment_end - 1:
                        # Populate the dictionary with the calculated positions
                        range_dict[contig_pos] = read_pos
            # Initialise a boolean of whether the current base passes filters, and should be added to the dictionary
            add_base = True
            # Iterate through the range_dict to extract the sequence of the bases in the range for both the read and
            # the reference sequence
            for contig_pos, read_pos in range_dict.items():
                try:
                    # Extract the sequence of the base
                    reference_base = reference_sequence[contig_pos]
                    base = read.alignment.query_sequence[read_pos]
                    # If any of the downstream or upstream bases do not match, set the boolean to False
                    if not match and reference_base != base:
                        add_base = False
                except KeyError:
                    pass
            # Populate the dictionary only if there are no other SNVs within five downstream and five upstream bases
            if add_base:
                unfiltered_read_details[read.alignment.qname][read.alignment.is_read1] = {
                    'mate unmapped': read.alignment.mate_is_unmapped,
                    'paired': read.alignment.is_paired,
                    'forward': read.alignment.is_read1,
                    'reverse': read.alignment.is_read2,
                    'match': match,
                    'qbase': query_base,
                    'rbase': ref_base,
                    'qual': quality,
                    'pos': column.pos,
                    'gene': column.reference_name
                }
                # Add the quality of the base
                if quality >= quality_cutoff:
                    qualities.append(quality)
    # Parse the unfiltered reads to characterise the bases
    for read_base, dir_dict in unfiltered_read_details.items():
        # Check to see if paired reads are present at this position
        if len(dir_dict) > 1:
            # SNV in both forward and reverse reads
            if not dir_dict[True]['match'] and not dir_dict[False]['match']:
                # Same SNV sequence - don't quality filter as the reads agreeing acts as a quality check
                if dir_dict[True]['qbase'] == dir_dict[False]['qbase']:
                    # Forward and reverse SNV - add two matches (for forward and reverse reads)
                    if dir_dict[True]['qbase'] not in filtered_read_dict['congruent_SNV']:
                        filtered_read_dict['congruent_SNV'][dir_dict[True]['qbase']] = 2
                    else:
                        filtered_read_dict['congruent_SNV'][dir_dict[True]['qbase']] += 2
                # Different SNV sequences
                else:
                    # Both SNV sequences pass quality
                    if dir_dict[True]['qual'] >= quality_cutoff and dir_dict[False]['qual'] >= quality_cutoff:
                        # Forward SNV1
                        if dir_dict[True]['qbase'] not in filtered_read_dict['forward_SNV_reverse_SNV1']:
                            filtered_read_dict['forward_SNV_reverse_SNV1'][dir_dict[True]['qbase']] = 1
                        else:
                            filtered_read_dict['forward_SNV_reverse_SNV1'][dir_dict[True]['qbase']] += 1
                        # Reverse SNV2
                        if dir_dict[False]['qbase'] not in filtered_read_dict['reverse_SNV_forward_SNV1']:
                            filtered_read_dict['reverse_SNV_forward_SNV1'][dir_dict[False]['qbase']] = 1
                        else:
                            filtered_read_dict['reverse_SNV_forward_SNV1'][dir_dict[False]['qbase']] += 1
                    # Only the forward reads pass quality
                    elif dir_dict[True]['qual'] >= quality_cutoff:
                        # Forward SNV reverse QF
                        if dir_dict[True]['qbase'] not in filtered_read_dict['forward_SNV_reverse_UM_QF']:
                            filtered_read_dict['forward_SNV_reverse_UM_QF'][dir_dict[True]['qbase']] = 1
                        else:
                            filtered_read_dict['forward_SNV_reverse_UM_QF'][dir_dict[True]['qbase']] += 1
                        # Reverse QF
                        if dir_dict[False]['qbase'] not in filtered_read_dict['reverse_quality_filtered']:
                            filtered_read_dict['reverse_quality_filtered'][dir_dict[False]['qbase']] = 1
                        else:
                            filtered_read_dict['reverse_quality_filtered'][dir_dict[False]['qbase']] += 1
                    # Only the reverse reads pass quality
                    elif dir_dict[False]['qual'] >= quality_cutoff:
                        # Reverse SNV forward QF
                        if dir_dict[False]['qbase'] not in filtered_read_dict['reverse_SNV_forward_UM_QF']:
                            filtered_read_dict['reverse_SNV_forward_UM_QF'][dir_dict[False]['qbase']] = 1
                        else:
                            filtered_read_dict['reverse_SNV_forward_UM_QF'][dir_dict[False]['qbase']] += 1
                        # Forward QF
                        if dir_dict[True]['qbase'] not in filtered_read_dict['forward_quality_filtered']:
                            filtered_read_dict['forward_quality_filtered'][dir_dict[True]['qbase']] = 1
                        else:
                            filtered_read_dict['forward_quality_filtered'][dir_dict[True]['qbase']] += 1
                    # Neither forward nor reverse reads pass quality
                    else:
                        # Forward QF
                        if dir_dict[True]['qbase'] not in filtered_read_dict['forward_quality_filtered']:
                            filtered_read_dict['forward_quality_filtered'][dir_dict[True]['qbase']] = 1
                        else:
                            filtered_read_dict['forward_quality_filtered'][dir_dict[True]['qbase']] += 1
                        # Reverse QF
                        if dir_dict[False]['qbase'] not in filtered_read_dict['reverse_quality_filtered']:
                            filtered_read_dict['reverse_quality_filtered'][dir_dict[False]['qbase']] = 1
                        else:
                            filtered_read_dict['reverse_quality_filtered'][dir_dict[False]['qbase']] += 1
            # SNV in forward read only
            elif not dir_dict[True]['match'] and dir_dict[False]['match']:
                # Since only the forward read supports the SNV, quality filter
                if dir_dict[True]['qual'] >= quality_cutoff:
                    if dir_dict[True]['qbase'] not in filtered_read_dict['forward_SNV_reverse_ref']:
                        filtered_read_dict['forward_SNV_reverse_ref'][dir_dict[True]['qbase']] = 1
                    else:
                        filtered_read_dict['forward_SNV_reverse_ref'][dir_dict[True]['qbase']] += 1
                    # Since the reverse base matches the reference, don't quality filter
                    if dir_dict[False]['qbase'] not in filtered_read_dict['forward_SNV_reverse_ref']:
                        filtered_read_dict['forward_SNV_reverse_ref'][dir_dict[False]['qbase']] = 1
                    else:
                        filtered_read_dict['forward_SNV_reverse_ref'][dir_dict[False]['qbase']] += 1
                # Reverse ref forward QF
                else:
                    if dir_dict[False]['qbase'] not in filtered_read_dict['reverse_ref_forward_UM_QF']:
                        filtered_read_dict['reverse_ref_forward_UM_QF'][dir_dict[False]['qbase']] = 1
                    else:
                        filtered_read_dict['reverse_ref_forward_UM_QF'][dir_dict[False]['qbase']] += 1
            # SNV in reverse read only
            elif dir_dict[True]['match'] and not dir_dict[False]['match']:
                # Quality filter
                if dir_dict[False]['qual'] >= quality_cutoff:
                    if dir_dict[False]['qbase'] not in filtered_read_dict['reverse_SNV_forward_ref']:
                        filtered_read_dict['reverse_SNV_forward_ref'][dir_dict[False]['qbase']] = 1
                    else:
                        filtered_read_dict['reverse_SNV_forward_ref'][dir_dict[False]['qbase']] += 1
                    # Since the forward base matches the reference, don't quality filter
                    if dir_dict[True]['qbase'] not in filtered_read_dict['reverse_SNV_forward_ref']:
                        filtered_read_dict['reverse_SNV_forward_ref'][dir_dict[True]['qbase']] = 1
                    else:
                        filtered_read_dict['reverse_SNV_forward_ref'][dir_dict[True]['qbase']] += 1
                # Forward ref reverse QF
                else:
                    if dir_dict[True]['qbase'] not in filtered_read_dict['forward_ref_reverse_UM_QF']:
                        filtered_read_dict['forward_ref_reverse_UM_QF'][dir_dict[True]['qbase']] = 1
                    else:
                        filtered_read_dict['forward_ref_reverse_UM_QF'][dir_dict[True]['qbase']] += 1
            # Both reads match the reference sequence - don't quality filter, and add two matches (for forward and
            # reverse reads)
            else:
                if dir_dict[True]['qbase'] not in filtered_read_dict['congruent_ref']:
                    filtered_read_dict['congruent_ref'][dir_dict[True]['qbase']] = 2
                else:
                    filtered_read_dict['congruent_ref'][dir_dict[True]['qbase']] += 2
        # Either the reads are unpaired, or only a single read aligns to this position on the gene (no overlap)
        else:
            for direction in dir_dict:
                # SNV supported by a single read
                if not dir_dict[direction]['match']:
                    if dir_dict[direction]['qual'] >= quality_cutoff:
                        # Forward
                        if direction:
                            if dir_dict[direction]['qbase'] not in filtered_read_dict['forward_SNV_reverse_UM_QF']:
                                filtered_read_dict['forward_SNV_reverse_UM_QF'][dir_dict[direction]['qbase']] = 1
                            else:
                                filtered_read_dict['forward_SNV_reverse_UM_QF'][dir_dict[direction]['qbase']] += 1
                        # Reverse
                        else:
                            if dir_dict[direction]['qbase'] not in filtered_read_dict['reverse_SNV_forward_UM_QF']:
                                filtered_read_dict['reverse_SNV_forward_UM_QF'][dir_dict[direction]['qbase']] = 1
                            else:
                                filtered_read_dict['reverse_SNV_forward_UM_QF'][dir_dict[direction]['qbase']] += 1
                    else:
                        # Forward QF
                        if dir_dict[direction]['qbase'] not in filtered_read_dict['forward_quality_filtered']:
                            filtered_read_dict['forward_quality_filtered'][dir_dict[direction]['qbase']] = 1
                        else:
                            filtered_read_dict['forward_quality_filtered'][dir_dict[direction]['qbase']] += 1

                # Match to the reference sequence supported by a single read
                else:
                    if dir_dict[direction]['qual'] >= quality_cutoff:
                        # Forward
                        if direction:
                            if dir_dict[direction]['qbase'] not in filtered_read_dict['forward_ref_reverse_UM_QF']:
                                filtered_read_dict['forward_ref_reverse_UM_QF'][dir_dict[direction]['qbase']] = 1
                            else:
                                filtered_read_dict['forward_ref_reverse_UM_QF'][dir_dict[direction]['qbase']] += 1
                        # Reverse
                        else:
                            if dir_dict[direction]['qbase'] not in filtered_read_dict['reverse_ref_forward_UM_QF']:
                                filtered_read_dict['reverse_ref_forward_UM_QF'][dir_dict[direction]['qbase']] = 1
                            else:
                                filtered_read_dict['reverse_ref_forward_UM_QF'][dir_dict[direction]['qbase']] += 1
                    else:
                        # Reverse QF
                        if dir_dict[direction]['qbase'] not in filtered_read_dict['reverse_quality_filtered']:
                            filtered_read_dict['reverse_quality_filtered'][dir_dict[direction]['qbase']] = 1
                        else:
                            filtered_read_dict['reverse_quality_filtered'][dir_dict[direction]['qbase']] += 1
    return filtered_read_dict, qualities


def determine_cutoff(qualities, reference_sequence, base_cutoff, error_cutoff=1.0):
    """
    Determine the base cutoff value to use for SNV determination
    :param qualities: List of phred scores for every base in the pileup
    :param reference_sequence: String of the FASTA reference gene sequence
    :param error_cutoff: Float of the error cutoff value to use. Default is 1.0%
    :return: base_cutoff: Int of calculated gene-specific cutoff value to use
    :return: The percentage error value corresponding to the base_cutoff value
    """
    depth_prob = 0
    # Only find the cutoff if there are qualities (the gene is present in the sample)
    if qualities:
        # Convert the gene-specific phred score to its probability e.g phred score of 20 is 1x10^(-20/10) = 1x10^(-2)
        # = 0.01
        phred_prob = 1 * 10 ** (-(mean(qualities) / 10))
        # Determine the probability that any base in the pileup will be incorrect. Multiply the phred probability by
        # the average depth of the sample (number of bases / length of the gene)
        depth_prob = phred_prob * len(qualities) / len(reference_sequence)
        # Find the lowest value of the base cutoff that gives a false positive error value under the error cutoff value
        # for the entire length of the gene. Take the depth_prob to the power of the current base_cutoff value.
        # Multiply this by the length of the gene to find find that value for the entire gene, and by 100 to convert
        # this to a percent value. e.g. 0.01^2 * 1674 * 100 = 16.74%. Note that this is above the default 1% threshold.
        # The base_cutoff will be incremented by one, and the calculation will be run again: 0.01^3 * 1674 * 100 =
        # 0.1674%, which is below the threshold, and would pass
        while depth_prob ** base_cutoff * len(reference_sequence) > error_cutoff * 100:
            base_cutoff += 1
    return base_cutoff, depth_prob ** base_cutoff * 100 * len(reference_sequence)


def find_multibase_positions(ref_base, filtered_read_dict, base_cutoff, base_fraction_cutoff):
    """

    :param ref_base: String of the sequence of the reference gene at the current position
    :param filtered_read_dict: Dictionary with of base types: read direction: count
    :param base_cutoff: Integer of the number of identical mismatches in a column required for a SNV call
    :param base_fraction_cutoff: Float fraction of bases necessary to support a SNV call
    :return: snv_dict: Dictionary of characterised bases of the current column e.g. total, number matching reference
    sequence, number of SNVs in both forward and reverse reads, etc.
    :return: passing_snv_dict: Dictionary summarising counts of categories of bases e.g. congruent, forward, reverse,
    paired
    :return: total_coverage: Integer of the total number of bases passing filter in the pileup
    """
    # Initialise a dictionary to store the counts of the characterised bases
    snv_dict = {
        'total': 0,
        'total_congruent': 0,
        'total_congruent_SNV': 0,
        'total_forward': 0,
        'total_forward_SNV': 0,
        'total_reverse': 0,
        'total_reverse_SNV': 0,
        'total_SNV': 0
    }
    # Initialise the total depth of the column to zero
    total_coverage = 0
    # Initialise a dictionary to store the count for each individual nucleotide at this position e.g. A:24, G:2
    base_count = dict()
    # Iterate through the categories e.g. congruent_ref in the dictionary
    for category, base_dict in filtered_read_dict.items():
        # Iterate through the sequence of each query base, and the corresponding count
        for base, count in base_dict.items():
            # Do not process the base if it has been flagged as 'filtered'
            if 'filtered' not in category:
                # Populate the base counting dictionary with the base sequence and increment the count
                if base not in base_count:
                    base_count[base] = count
                else:
                    base_count[base] += count
                # Update the total coverage
                total_coverage += count
                snv_dict['total'] += count
            # Forward and reverse reads agree on base
            if 'congruent' in category:
                # Congruent reference sequence
                snv_dict['total_congruent'] += count
                snv_dict['total_forward'] += int(count / 2)
                snv_dict['total_reverse'] += int(count / 2)
                # Congruent SNVs
                if 'SNV' in category and base != ref_base:
                    snv_dict['total_congruent_SNV'] += count
                    snv_dict['total_forward_SNV'] += int(count / 2)
                    snv_dict['total_reverse_SNV'] += int(count / 2)
                    snv_dict['total_SNV'] += count
            # SNV in forward read
            elif category.startswith('forward_SNV'):
                snv_dict['total_forward'] += count
                if base != ref_base:
                    snv_dict['total_forward_SNV'] += count
                    snv_dict['total_SNV'] += count
            # Forward read matches reference
            elif category.startswith('forward_ref'):
                snv_dict['total_forward'] += count
            # SNV in reverse read
            elif category.startswith('reverse_SNV'):
                snv_dict['total_reverse'] += count
                if base != ref_base:
                    snv_dict['total_reverse_SNV'] += count
                    snv_dict['total_SNV'] += count
            # Reverse read match reference
            elif category.startswith('reverse_ref'):
                snv_dict['total_reverse'] += count
    # Initialise a dictionary to store the summary of characterised base types
    passing_snv_dict = {
        'congruent': dict(),
        'forward': dict(),
        'reverse': dict(),
        'paired': dict()
    }
    # Boolean of whether there are bases passing filter, and the passing_snv_dict should be used
    return_dict = False
    # Iterate through the categories in the dictionary
    for category, base_dict in filtered_read_dict.items():
        # Iterate through each query base and its corresponding count
        for base, count in base_dict.items():
            # Congruent SNVs
            if 'congruent' in category and base != ref_base:
                # If the base_cutoff_fraction has been provided, use it in making SNV calls
                if base_fraction_cutoff:
                    # Ensure that the number of SNVs in the pileup is greater than the base_cutoff and the
                    # fraction of SNVs in the pileup is greater than the base_cutoff_fraction
                    if float(base_count[base] / snv_dict['total']) >= base_fraction_cutoff and base_count[base] \
                            >= base_cutoff:
                        # Update the summary dictionary with the base sequence and count
                        if base not in passing_snv_dict['forward']:
                            passing_snv_dict['congruent'][base] = count
                        else:
                            passing_snv_dict['congruent'][base] += count
                        return_dict = True
                # If base_cutoff_fraction is not supplied, only the number of SNVs in the pileup must be greater than
                # the base_cutoff value in order for a SNV call
                else:
                    if base_count[base] >= base_cutoff:
                        if base not in passing_snv_dict['congruent']:
                            passing_snv_dict['congruent'][base] = count
                        else:
                            passing_snv_dict['congruent'][base] += count
                    return_dict = True
            # SNVs in the forward read
            if category.startswith('forward_SNV') and base != ref_base:
                if base_fraction_cutoff:
                    if float(base_count[base] / snv_dict['total']) >= base_fraction_cutoff and base_count[base] \
                            >= base_cutoff:
                        if base not in passing_snv_dict['forward']:
                            passing_snv_dict['forward'][base] = count
                        else:
                            passing_snv_dict['forward'][base] += count
                        return_dict = True
                else:
                    if base_count[base] >= base_cutoff:
                        if base not in passing_snv_dict['forward']:
                            passing_snv_dict['forward'][base] = count
                        else:
                            passing_snv_dict['forward'][base] += count
                        return_dict = True
            # SNVs in the reverse read
            if category.startswith('reverse_SNV') and base != ref_base:
                if base_fraction_cutoff:
                    if float(base_count[base] / snv_dict['total']) >= base_fraction_cutoff and base_count[base] \
                            >= base_cutoff:
                        if base not in passing_snv_dict['reverse']:
                            passing_snv_dict['reverse'][base] = count
                        else:
                            passing_snv_dict['reverse'][base] += count
                        return_dict = True
                else:
                    if base_count[base] >= base_cutoff:
                        if base not in passing_snv_dict['reverse']:
                            passing_snv_dict['reverse'][base] = count
                        else:
                            passing_snv_dict['reverse'][base] += count
                        return_dict = True
            # All unfiltered bases
            if base != ref_base and 'filtered' not in category:
                if base_fraction_cutoff:
                    if float(base_count[base] / snv_dict['total']) >= base_fraction_cutoff and base_count[base] \
                            >= base_cutoff:
                        if base not in passing_snv_dict['paired']:
                            passing_snv_dict['paired'][base] = count
                        else:
                            passing_snv_dict['paired'][base] += count
                        return_dict = True
                else:
                    if base_count[base] >= base_cutoff:
                        if base not in passing_snv_dict['paired']:
                            passing_snv_dict['paired'][base] = count
                        else:
                            passing_snv_dict['paired'][base] += count
                        return_dict = True
    if return_dict:
        return snv_dict, passing_snv_dict, total_coverage
    else:
        return snv_dict, dict(), total_coverage


def position_details(actual_position, passing_snv_dict, contig_name, ref_base, total_coverage, base_cutoff, error_perc):
    """
    Create a string summarising the ConFindr results for current gene
    :param actual_position: Integer of the adjusted position of the current base
    :param passing_snv_dict: Dictionary summarising counts of categories of bases
    :param contig_name: Name of current gene
    :param ref_base: Sequence of the current reference base
    :param total_coverage: Integer of the total depth at this position
    :param base_cutoff: Integer of the SNV depth cutoff value
    :param error_perc: Float of the calculated error value
    :return: to_write: String containing formatted outputs to be entered into the final report
    """
    # List of the base categories present in passing_snv_dict
    read_types = ['congruent', 'paired', 'forward', 'reverse']
    # Update the to_write string with basic information
    to_write = '{contig_name},{pos},{ref_base},'.format(contig_name=contig_name,
                                                        pos=actual_position,
                                                        ref_base=ref_base)
    # Initialise the coverage value to zero
    snv_coverage = 0
    # Iterate through all the read types
    for read_type in read_types:
        # Boolean of whether a semi-colon needs to be added to the string to separate all the base sequences
        semi_colon = False
        for base, coverage in passing_snv_dict[read_type].items():
            # Ensure that the base isn't empty
            if base:
                if semi_colon:
                    to_write += ';'
                to_write += '{base}:{depth}'.format(base=base,
                                                    depth=coverage)
                semi_colon = True
                # Increment the coverage only if the read type isn't 'paired'
                if read_type != 'paired':
                    snv_coverage += coverage
        to_write += ','
    # Format the error_perc to be more readable
    error_perc = '{:0.2f}'.format(error_perc) if error_perc else 'ND'
    to_write += '{snv_cov},{total_cov},{base_cutoff},{error_perc}\n'\
        .format(snv_cov=snv_coverage,
                total_cov=total_coverage,
                base_cutoff=base_cutoff,
                error_perc=error_perc)
    return to_write


def read_contig(contig_name, bamfile_name, reference_fasta, allele_records, 
                fastq_records, quality_cutoff=20, base_cutoff=None, 
                base_fraction_cutoff=None, fasta=False, error_cutoff=1.0,
                nanopore=False):
    """
    Examines a contig to find if there are positions where more than one base is present.
    :param contig_name: Name of contig as a string.
    :param bamfile_name: Full path to bamfile. Must be sorted/indexed
    :param reference_fasta: Full path to fasta file that was used to generate the bamfile.
    :param allele_records:
    :param fastq_records: Dictionary of SeqIO records parsed from filtered FASTQ reads
    :param quality_cutoff: Bases must have at least this phred score to be considered (INT)
    :param base_cutoff: At least this many bases must support a minor variant (INT)
    :param base_fraction_cutoff: At least this percentage of bases must support minor variant (FLOAT)
    :param fasta: Boolean on whether the samples are in FASTA format. Default is False
    :param error_cutoff: Float of the error cutoff value to use. Default is 1.0%
    :param nanopore: Boolean of whether Nanopore reads were provided. Default is False.
    :return: Dictionary of positions where more than one base is present. Keys are contig name, values are positions
    """
    pysam_fasta = pysam.FastaFile(reference_fasta)
    multibase_position_dict = dict()
    to_write = str()
    # If analysing FASTA files, a single base difference is all that is expected
    if fasta:
        base_cutoff = 1
    reference_sequence = str(allele_records[contig_name].seq)
    # Parse the BAM file with pysam to create AlignmentFile, and AlignmentFile.pileup objects
    bamfile, pileup = parse_bam(bamfile_name=bamfile_name,
                                contig_name=contig_name,
                                pysam_fasta=pysam_fasta)
    filtered_read_dict = dict()
    quality_list = list()
    for i, column in enumerate(pileup):
        filtered_reads, qualities = characterise_read(column=column,
                                                      reference_sequence=reference_sequence,
                                                      fastq_records=fastq_records,
                                                      quality_cutoff=quality_cutoff,
                                                      fasta=fasta,
                                                      nanopore=nanopore)
        filtered_read_dict[i] = filtered_reads
        quality_list += qualities
    # Initialise the calculated error percentage to zero
    error_perc = None
    # If the base_cutoff is not manually specified on the command-line, adjust it based
    # upon overall sequence quality
    if base_cutoff != 3:
        base_cutoff, error_perc = determine_cutoff(qualities=quality_list,
                                                   reference_sequence=reference_sequence,
                                                   base_cutoff = base_cutoff,
                                                   error_cutoff=error_cutoff)
    bamfile.close()
    # It seems that the pileup (generator?) is used up above, so it must be recreated
    bamfile, pileup = parse_bam(bamfile_name=bamfile_name,
                                contig_name=contig_name,
                                pysam_fasta=pysam_fasta)
    for i, column in enumerate(pileup):
        # Extract the sequence of the reference gene at the current position
        ref_base = reference_sequence[column.pos]
        # Summarise the pileup
        snv_dict, passing_snv_dict, total_coverage = \
            find_multibase_positions(ref_base=ref_base,
                                     filtered_read_dict=filtered_read_dict[i],
                                     base_cutoff=base_cutoff,
                                     base_fraction_cutoff=base_fraction_cutoff)
        # If there are any SNVs called for the gene, update the multibase_position_dict, and to_write string
        if passing_snv_dict:
            # Pysam starts counting at 0, whereas we actually want to start counting at 1.
            actual_position = column.pos + 1
            # Initialise the gene name in the dictionary as required
            if column.reference_name not in multibase_position_dict:
                multibase_position_dict[column.reference_name] = dict()
            # Update the dictionary with the actual position:
            multibase_position_dict[column.reference_name].update({actual_position: passing_snv_dict})
            to_write += position_details(actual_position=actual_position,
                                         passing_snv_dict=passing_snv_dict,
                                         contig_name=contig_name,
                                         ref_base=ref_base,
                                         total_coverage=total_coverage,
                                         base_cutoff=base_cutoff,
                                         error_perc=error_perc)
    bamfile.close()
    return multibase_position_dict, to_write


def find_rmlst_type(kma_report, rmlst_report):
    """
    Uses a report generated by KMA to determine what allele is present for each rMLST gene.
    :param kma_report: The .res report generated by KMA.
    :param rmlst_report: rMLST report file to write information to.
    :return: a sorted list of loci present, in format 'gene_allele'
    """
    genes_to_use = dict()
    score_dict = dict()
    gene_alleles = list()
    with open(kma_report) as tsvfile:
        reader = csv.DictReader(tsvfile, delimiter='\t')
        for row in reader:
            gene_allele = row['#Template']
            score = int(row['Score'])
            gene = gene_allele.split('_')[0]
            allele = gene_allele.split('_')[1]
            if gene not in score_dict:
                score_dict[gene] = score
                genes_to_use[gene] = allele
            else:
                if score > score_dict[gene]:
                    score_dict[gene] = score
                    genes_to_use[gene] = allele
    for gene in genes_to_use:
        gene_alleles.append(gene + '_' + genes_to_use[gene].replace(' ', ''))
    gene_alleles = sorted(gene_alleles)
    with open(rmlst_report, 'w') as f:
        f.write('Gene,Allele\n')
        for gene_allele in gene_alleles:
            gene = gene_allele.split('_')[0]
            allele = gene_allele.split('_')[1]
            f.write('{},{}\n'.format(gene, allele))
    return gene_alleles


def base_dict_to_string(base_dict):
    """
    Converts a dictionary to a string. {'C': 12, 'A':4} gets converted to C:12;A:4
    :param base_dict: Dictionary of bases and counts created by find_if_multibase
    :return: String representing that dictionary.
    """
    outstr = ''
    # First, sort base_dict so that major allele always comes first - makes output report nicer to look at.
    base_list = sorted(base_dict.items(), key=lambda kv: kv[1], reverse=True)
    for base in base_list:
        outstr += '{}:{};'.format(base[0], base[1])
    return outstr[:-1]


def find_total_sequence_length(fasta_file):
    """
    Totals up number of bases in a fasta file.
    :param fasta_file: Path to an uncompressed, fasta-formatted file.
    :return: Number of total bases in file, as an int.
    """
    total_length = 0
    for sequence in SeqIO.parse(fasta_file, 'fasta'):
        total_length += len(sequence.seq)
    return total_length


# Removing percentage contamination estimation until it can be made more 
# accurate
#-------------------------------------------------------------------------------
# def estimate_percent_contamination(contamination_report_file):
#     """
#     Estimates the percent contamination of a sample (and standard deviation).
#     :param contamination_report_file: File created by read_contig,
#     :return: Estimated percent contamination and standard deviation.
#     """
#     # Initialise a list to store the percentage of SNVs out of the total bases
#     contam_levels = list()
#     with open(contamination_report_file) as csvfile:
#         reader = csv.DictReader(csvfile)
#         for row in reader:
#             base_counts = int(row['SNVCoverage'])
#             total_coverage = int(row['TotalCoverage'])
#             contam_levels.append(base_counts * 100 / total_coverage)
#     return '%.2f' % (np.mean(contam_levels)), '%.2f' % np.std(contam_levels)


def load_fastq_records(gz, paired, forward):
    """
    Use SeqIO to load FASTQ records from file
    :param gz: Name and path of FASTQ file
    :param paired: Boolean of whether the reads are paired
    :param forward: Boolean of whether the current reads are in the forward direction
    :return: records: Dictionary of SeqIO parsed FASTQ records with consistent ID naming
    """
    # Initialise a dictionary to store the FASTQ records
    records = dict()
    # Iterate through the reads
    for record in SeqIO.parse(gz, 'fastq'):
        # Only update the naming scheme for paired reads
        if paired:
            if forward:
                # Change a :1: to /1 in the record.id
                if ':1:' in record.id:
                    record.id = record.id + '/1'
                # Don't worry if the record.id already has a /1
                elif '/1' in record.id:
                    pass
                # If the record.id doesn't have a read direction, add /1
                else:
                    record.id = record.id + '/1'
            # Process reverse reads in a similar fashion to forward reads
            else:
                if ':2:' in record.id:
                    record.id = record.id + '/2'
                elif '/2' in record.id:
                    pass
                else:
                    record.id = record.id + '/2'
        else:
            pass
        records.update(SeqIO.to_dict([record]))
    return records


def index_databases(sample_database):
    """
    Index the database file with pysam and kma
    :param sample_database:
    """
    if not os.path.isfile(sample_database + '.fai'):  # Don't bother re-indexing, this only needs to happen once.
        try:
            pysam.faidx(sample_database)
        except pysam.utils.SamtoolsError:
            pass
    kma_database = sample_database.replace('.fasta', '') + '_kma'

    if not os.path.isfile(kma_database + '.name'):  # The .name is one of the files KMA creates when making a database.
        logging.info('Since this is the first time you are using this database, it needs to be indexed by KMA. '
                     'This might take a while')
        cmd = 'kma index -i {} -o {}'.format(sample_database, kma_database)  # NOTE: Need KMA >=1.2.0 for this to work.
        out = str()
        try:
            out, err = run_cmd(cmd)
        except subprocess.CalledProcessError as e:
            err = e
        log = sample_database + '_log.txt'
        write_to_logfile(log, out, err, cmd)


# noinspection PyUnresolvedReferences
def find_contamination(pair, output_folder, databases_folder, base_cutoff, forward_id='_R1', threads=1,
                       keep_files=False,
                       quality_cutoff=20, base_fraction_cutoff=0.05, cgmlst_db=None, xmx=None,
                       tmpdir=None, data_type='Illumina', use_rmlst=False, min_matching_hashes=40,
                       fasta=False, error_cutoff=1.0, debug=False):
    """
    This needs some documentation fairly badly, so here we go.
    :param pair: This has become a misnomer. If the input reads are actually paired, needs to be a list
    with the full filepath to forward reads at index 0 and full path to reverse reads at index 1.
    If reads are unpaired, should be a list of length 1 with the only entry being the full filepath to read set.
    :param output_folder: Folder where outputs (confindr log and report, and other stuff) will be stored.
    This will be created if it does not exist. (I think - should write a test that double checks this).
    :param databases_folder: Full path to folder where ConFindr's databases live. These files can be
    downloaded from figshare in .tar.gz format (https://ndownloader.figshare.com/files/11864267), and
    will be automatically downloaded if the script is run from the command line.
    :param forward_id: Identifier that marks reads as being in the forward direction for paired reads.
    Defaults to _R1
    :param threads: Number of threads to run analyses with. All parts of this pipeline scale pretty well,
    so more is better.
    :param keep_files: Boolean that says whether or not to keep temporary files.
    :param quality_cutoff: Integer of the phred score required to have a base count towards a multiallelic site.
    :param base_cutoff: Integer of number of bases needed to have a base be part of a multiallelic site.
    :param base_fraction_cutoff: Float of fraction of bases needed to have a base be part of a multiallelic site.
    If specified will be used in parallel with base_cutoff
    :param cgmlst_db: if None, we're using rMLST, if True, using some sort of custom cgMLST database. This requires some
    custom parameters.
    :param xmx: if None, BBTools will use auto memory detection. If string, BBTools will use what's specified as their
    memory request.
    :param tmpdir: if None, any genus-specifc databases that need to be created will be written to ConFindr DB location.
    :param data_type: Either Illumina or Nanopore, depending on what type your reads are. (STR)
    :param use_rmlst: If False, use cgderived data instead of rMLST where possible. If True, always use rMLST. (BOOL)
    :param min_matching_hashes: Minimum number of matching hashes in a MASH screen in order for a genus to be
    considered present in a sample. Default is 40
    :param fasta: Boolean on whether the samples are in FASTA format. Default is False
    :param error_cutoff: Float of the error cutoff value to use. Default is 1.0%
    :param debug: Run the find_contamination with multi-processing to allow easier debugging
    """
    if os.path.isfile(os.path.join(databases_folder, 'download_date.txt')):
        with open(os.path.join(databases_folder, 'download_date.txt')) as f:
            database_download_date = f.readline().rstrip()
    else:
        database_download_date = 'ND'
    log = os.path.join(output_folder, 'confindr_log.txt')
    if len(pair) == 2:
        sample_name = os.path.split(pair[0])[-1].split(forward_id)[0]
        paired = True
        logging.debug('Sample is paired. Sample name is {}'.format(sample_name))
    else:
        sample_name = os.path.split(pair[0])[-1].split('.')[0]
        paired = False
        logging.debug('Sample is unpaired. Sample name is {}'.format(sample_name))
    sample_tmp_dir = os.path.join(output_folder, sample_name)
    if not os.path.isdir(sample_tmp_dir):
        os.makedirs(sample_tmp_dir)

    logging.info('Checking for cross-species contamination...')
    if paired:
        genus = find_cross_contamination(databases_folder,
                                         reads=pair,
                                         sample_name=sample_name,
                                         tmpdir=sample_tmp_dir,
                                         log=log,
                                         threads=threads,
                                         min_matching_hashes=min_matching_hashes)
    else:
        genus = find_cross_contamination(databases_folder,
                                         reads=pair[0],
                                         sample_name=sample_name,
                                         tmpdir=sample_tmp_dir,
                                         log=log,
                                         threads=threads,
                                         min_matching_hashes=min_matching_hashes)
    # if len(genus.split(':')) > 1:
    #     write_output(output_report=os.path.join(output_folder, 'confindr_report.csv'),
    #                     sample_name=sample_name,
    #                     multi_positions=0,
    #                     genus=genus,
    #                     total_gene_length=0,
    #                     database_download_date=database_download_date)
    #     logging.info('Found cross-contamination! Skipping rest of analysis...\n')
    #     if keep_files is False:
    #         shutil.rmtree(sample_tmp_dir)
    #     return
    # Setup genus-specific databases, if necessary.
    if cgmlst_db is not None:
        # Sanity check that the DB specified is actually a file, otherwise, quit with appropriate error message.
        if not os.path.isfile(cgmlst_db):
            logging.error('ERROR: Specified cgMLST file ({}) does not exist. Please check the path and try again.'
                          .format(cgmlst_db))
            quit(code=1)
        sample_database = cgmlst_db
    else:
        db_folder = databases_folder if tmpdir is None else tmpdir
        if not os.path.isdir(db_folder):
            os.makedirs(db_folder)
        if genus != 'ND':
            # Logic here is as follows: users can either have both rMLST databases, which cover all of bacteria,
            # cgmlst-derived databases, which cover only Escherichia, Salmonella, and Listeria (may add more at some
            # point), or they can have both. They can also set priority to either always use rMLST, or to use my
            # core-genome derived stuff and fall back on rMLST if they're trying to look at a genus I haven't created
            # a scheme for.
            #
            if len(genus.split(':')) > 1:
                predominant_genus = genus.split(':')[0]
            else:
                predominant_genus = genus
            # In the event rmlst databases have priority, always use them.
            if use_rmlst is True:
                sample_database = os.path.join(db_folder, '{}_db.fasta'.format(predominant_genus))
                if not os.path.isfile(sample_database):

                    if os.path.isfile(os.path.join(db_folder, 'gene_allele.txt')) and \
                            os.path.isfile(os.path.join(db_folder, 'rMLST_combined.fasta')):
                        logging.info('Setting up rMLST genus-specific database for genus {}...'
                                     .format(predominant_genus))
                        allele_list = find_genus_specific_allele_list(os.path.join(db_folder, 'gene_allele.txt'),
                                                                      predominant_genus)
                        # Create the allele-specific database
                        setup_allelespecific_database(fasta_file=sample_database,
                                                      database_folder=db_folder,
                                                      allele_list=allele_list)
            else:
                # Check if a cgderived database is available. If not, try to use rMLST database.
                sample_database = os.path.join(db_folder, '{}_db_cgderived.fasta'.format(predominant_genus))
                if not os.path.isfile(sample_database):
                    sample_database = os.path.join(db_folder, '{}_db.fasta'.format(predominant_genus))
                    # Create genus specific database if it doesn't already exist and we have the necessary rMLST files.
                    if os.path.isfile(os.path.join(db_folder, 'rMLST_combined.fasta')) and \
                            os.path.isfile(os.path.join(db_folder, 'gene_allele.txt')) and not \
                            os.path.isfile(sample_database):
                        logging.info('Setting up core genome genus-specific database for genus {}...'
                                     .format(predominant_genus))
                        allele_list = find_genus_specific_allele_list(os.path.join(db_folder, 'gene_allele.txt'),
                                                                      predominant_genus)
                        setup_allelespecific_database(fasta_file=sample_database,
                                                      database_folder=db_folder,
                                                      allele_list=allele_list)

        else:
            sample_database = os.path.join(db_folder, 'rMLST_combined.fasta')

    # If a user has gotten to this point and they don't have any database available to do analysis because
    # they don't have rMLST downloaded and we don't have a cg-derived database available, boot them with a helpful
    # message.
    if not os.path.isfile(sample_database):
        write_output(output_report=os.path.join(output_folder, 'confindr_report.csv'),
                     sample_name=sample_name,
                     multi_positions=0,
                     genus=genus,
                     total_gene_length=0,
                     database_download_date=database_download_date)
        logging.info('Did not find databases for genus {genus}. You can download the rMLST database to get access to '
                     'all genera (see https://olc-bioinformatics.github.io/ConFindr/install/). Alternatively, if you '
                     'have a high-quality core-genome derived database for your genome of interest, we would be happy '
                     'to add it - open an issue at https://github.com/OLC-Bioinformatics/ConFindr/issues with the '
                     'title "Add genus-specific database: {genus}"\n'.format(genus=genus))
        if keep_files is False:
            shutil.rmtree(sample_tmp_dir)
        return

    # Extract rMLST reads and quality trim.
    logging.info('Extracting conserved core genes...')
    out, err, cmd = '', '', ''
    forward_bait = str()
    forward_trimmed = str()
    reverse_bait = str()
    reverse_trimmed = str()
    unpaired_bait = str()
    unpaired_trimmed = str()
    if paired:
        forward_bait = os.path.join(sample_tmp_dir, '{sn}_baited_R1.fastq.gz'.format(sn=sample_name))
        reverse_bait = forward_bait.replace('_R1', '_R2')
        if not os.path.isfile(forward_bait):
            if xmx is None:
                out, err, cmd = bbtools.bbduk_bait(reference=sample_database,
                                                   forward_in=pair[0],
                                                   reverse_in=pair[1],
                                                   forward_out=forward_bait,
                                                   reverse_out=reverse_bait,
                                                   threads=threads,
                                                   returncmd=True)
            else:
                out, err, cmd = bbtools.bbduk_bait(reference=sample_database,
                                                   forward_in=pair[0],
                                                   reverse_in=pair[1],
                                                   forward_out=forward_bait,
                                                   reverse_out=reverse_bait,
                                                   threads=threads,
                                                   Xmx=xmx,
                                                   returncmd=True)
    else:
        # Still name the file '_baited_trimmed' even if the file won't be trimmed
        if data_type == 'Nanopore' or fasta:
            unpaired_bait = os.path.join(sample_tmp_dir, '{sn}_baited_trimmed.fastq.gz'.format(sn=sample_name))
        else:
            unpaired_bait = os.path.join(sample_tmp_dir, '{sn}_baited.fastq.gz'.format(sn=sample_name))
        if not os.path.isfile(unpaired_bait):
            if xmx is None:
                out, err, cmd = bbtools.bbduk_bait(reference=sample_database,
                                                   forward_in=pair[0],
                                                   forward_out=unpaired_bait,
                                                   returncmd=True, threads=threads)
            else:
                out, err, cmd = bbtools.bbduk_bait(reference=sample_database,
                                                   forward_in=pair[0],
                                                   forward_out=unpaired_bait,
                                                   Xmx=xmx,
                                                   returncmd=True, threads=threads)
    if out:
        write_to_logfile(log, out, err, cmd)
    logging.info('Quality trimming...')
    out, err, cmd = '', '', ''
    if data_type == 'Illumina':
        if paired:
            forward_trimmed = os.path.join(sample_tmp_dir, '{sn}_baited_trimmed_R1.fastq.gz'.format(sn=sample_name))
            reverse_trimmed = forward_trimmed.replace('_R1', '_R2')
            if not os.path.isfile(forward_trimmed):
                if xmx is None:
                    out, err, cmd = bbtools.bbduk_trim(forward_in=forward_bait,
                                                       reverse_in=reverse_bait,
                                                       forward_out=forward_trimmed,
                                                       reverse_out=reverse_trimmed,
                                                       threads=str(threads), returncmd=True)
                else:
                    out, err, cmd = bbtools.bbduk_trim(forward_in=forward_bait,
                                                       reverse_in=reverse_bait,
                                                       forward_out=forward_trimmed,
                                                       reverse_out=reverse_trimmed,
                                                       Xmx=xmx,
                                                       threads=str(threads),
                                                       returncmd=True)

            with gzip.open(forward_trimmed, 'rt') as gz:
                fastq_records = load_fastq_records(gz=gz,
                                                   paired=True,
                                                   forward=True)
            with gzip.open(reverse_trimmed, 'rt') as gz:
                # fastq_records.update(SeqIO.to_dict(SeqIO.parse(gz, 'fastq')))
                fastq_records.update(load_fastq_records(gz=gz,
                                                        paired=True,
                                                        forward=False))

        else:
            unpaired_trimmed = os.path.join(sample_tmp_dir, '{sn}_baited_trimmed.fastq.gz'.format(sn=sample_name))

            if not fasta:
                if not os.path.isfile(unpaired_trimmed):
                    if xmx is None:
                        out, err, cmd = bbtools.bbduk_trim(forward_in=unpaired_bait,
                                                           forward_out=unpaired_trimmed,
                                                           returncmd=True,
                                                           threads=threads)
                    else:
                        out, err, cmd = bbtools.bbduk_trim(forward_in=unpaired_bait,
                                                           forward_out=unpaired_trimmed,
                                                           returncmd=True,
                                                           threads=threads,
                                                           Xmx=xmx)
                with gzip.open(unpaired_trimmed, 'rt') as gz:
                    # fastq_records = SeqIO.to_dict(SeqIO.parse(gz, 'fastq'))
                    fastq_records = load_fastq_records(gz=gz,
                                                       paired=False,
                                                       forward=True)
            else:
                # '''unpaired_bait'''
                with gzip.open(unpaired_bait, 'rt') as gz:
                    # fastq_records = SeqIO.to_dict(SeqIO.parse(gz, 'fastq'))
                    fastq_records = load_fastq_records(gz=gz,
                                                       paired=False,
                                                       forward=True)
        write_to_logfile(log, out, err, cmd)
    else:
        if paired:
            with gzip.open(forward_bait, 'rt') as gz:
                # fastq_records = SeqIO.to_dict(SeqIO.parse(gz, 'fastq'))
                fastq_records = load_fastq_records(gz=gz,
                                                   paired=True,
                                                   forward=True)
            with gzip.open(reverse_bait, 'rt') as gz:
                # fastq_records.update(SeqIO.to_dict(SeqIO.parse(gz, 'fastq')))
                fastq_records.update(load_fastq_records(gz=gz,
                                                        paired=True,
                                                        forward=False))
        else:
            with gzip.open(unpaired_bait, 'rt') as gz:
                # fastq_records = SeqIO.to_dict(SeqIO.parse(gz, 'fastq'))
                fastq_records = load_fastq_records(gz=gz,
                                                   paired=False,
                                                   forward=True)
    logging.info('Detecting contamination...')
    # Now do mapping in two steps - first, map reads back to database with ambiguous reads matching all - this
    # will be used to get a count of number of reads aligned to each gene/allele so we can create a custom rmlst file
    # with only the most likely allele for each gene.
    kma_report = os.path.join(sample_tmp_dir, '{sn}_kma'.format(sn=sample_name))
    # if not os.path.isfile(sample_database + '.fai'):  # Don't bother re-indexing, this only needs to happen once.
    #     pysam.faidx(sample_database)
    kma_database = sample_database.replace('.fasta', '') + '_kma'
    #
    # if not os.path.isfile(kma_database + '.name'):  # The .name is one of the files KMA creates when making a database
    #     logging.info('Since this is the first time you are using this database, it needs to be indexed by KMA. '
    #                  'This might take a while')
    #     cmd = 'kma index -i {} -o {}'.format(sample_database, kma_database)  # NOTE: Need KMA >=1.2.0 for this to work
    #     out, err = run_cmd(cmd)
    #     write_to_logfile(log, out, err, cmd)
    index_databases(sample_database=sample_database)
    # Run KMA.
    if paired:
        if not os.path.isfile(kma_report + '.res'):
            cmd = 'kma -ipe {forward_in} {reverse_in} -t_db {kma_database} -o {kma_report} ' \
                  '-t {threads}'.format(forward_in=forward_trimmed,
                                        reverse_in=reverse_trimmed,
                                        kma_database=kma_database,
                                        kma_report=kma_report,
                                        threads=threads)
            out, err = run_cmd(cmd)
            write_to_logfile(log, out, err, cmd)
    else:
        if not os.path.isfile(kma_report + '.res'):
            if data_type == 'Illumina':
                # Use the FASTA file (rather than the reads) as the input
                if fasta:
                    cmd = 'kma -i {input_reads} -t_db {kma_database} -mem_mode -ID 100 -ConClave 2 -ex_mode ' \
                          '-o {kma_report} -t {threads}' \
                        .format(input_reads=pair[0],
                                kma_database=kma_database,
                                kma_report=kma_report,
                                threads=threads)
                else:
                    cmd = 'kma -i {input_reads} -t_db {kma_database} -o {kma_report} ' \
                          '-t {threads}'.format(input_reads=unpaired_trimmed,
                                                kma_database=kma_database,
                                                kma_report=kma_report,
                                                threads=threads)
            else:
                # Recommended Nanopore settings from KMA repo: https://bitbucket.org/genomicepidemiology/kma
                cmd = 'kma -i {input_reads} -t_db {kma_database} -o {kma_report} -mem_mode -mp 20 -mrs 0.0 -bcNano ' \
                      '-t {threads}'.format(input_reads=unpaired_bait,
                                            kma_database=kma_database,
                                            kma_report=kma_report,
                                            threads=threads)
            out, err = run_cmd(cmd)
            write_to_logfile(log, out, err, cmd)

    rmlst_report = os.path.join(output_folder, sample_name + '_alleles.csv')
    gene_alleles = find_rmlst_type(kma_report=kma_report + '.res',
                                   rmlst_report=rmlst_report)

    rmlst_fasta = os.path.join(sample_tmp_dir, '{sn}_alleles.fasta'.format(sn=sample_name))
    if not os.path.isfile(rmlst_fasta):
        with open(rmlst_fasta, 'w') as f:
            for contig in SeqIO.parse(sample_database, 'fasta'):
                if contig.id in gene_alleles:
                    SeqIO.write(contig, f, 'fasta')
                    # f.write('>{}\n'.format(contig.id))
                    # f.write(str(contig.seq) + '\n')
    rmlst_gene_length = find_total_sequence_length(rmlst_fasta)
    logging.debug('Total gene length is {}'.format(rmlst_gene_length))
    pysam_pass = True
    # Second step of mapping - Do a mapping of our baited reads against a fasta file that has only one allele per
    # rMLST gene.
    try:
        if not os.path.isfile(rmlst_fasta + '.fai'):
            pysam.faidx(rmlst_fasta)
        outbam = os.path.join(sample_tmp_dir, '{sn}_contamination.bam'.format(sn=sample_name))
        sorted_bam = os.path.join(sample_tmp_dir, '{sn}_contamination_sorted.bam'.format(sn=sample_name))
        outsam = outbam.replace('.bam', '.sam')
        if not os.path.isfile(sorted_bam):
            if paired:
                cmd = 'bbmap.sh ref={ref} in={forward_in} in2={reverse_in} out={outbam} threads={threads} mdtag ' \
                      'nodisk'.format(ref=rmlst_fasta,
                                      forward_in=forward_trimmed,
                                      reverse_in=reverse_trimmed,
                                      outbam=outbam,
                                      threads=threads)
                if cgmlst_db is not None:
                    # Lots of core genes seem to have relatives within a genome that are at ~70% identity. This means
                    # that reads that shouldn't map do, and cause false positives. Adding in this sub-filter means that
                    # reads can only have one mismatch, so they have to be from the right gene for this to work.
                    cmd += ' subfilter=1'
                if xmx:
                    cmd += ' -Xmx{}'.format(xmx)
                out, err = run_cmd(cmd)
                write_to_logfile(log, out, err, cmd)
            else:
                #
                if data_type == 'Illumina' and not fasta:
                    cmd = 'bbmap.sh ref={ref} in={forward_in} out={outbam} threads={threads} mdtag ' \
                          'nodisk'.format(ref=rmlst_fasta,
                                          forward_in=unpaired_trimmed,
                                          outbam=outbam,
                                          threads=threads)
                    if cgmlst_db is not None:
                        # Core genes can have relatives within a genome that are at ~70 percent identity. This means
                        # that reads that shouldn't map do, and cause false positives. Adding in this sub-filter means
                        # reads can only have one mismatch, so they have to be from the right gene for this to work.
                        cmd += ' subfilter=1'
                    if xmx:
                        cmd += ' -Xmx{}'.format(xmx)
                    out, err = run_cmd(cmd)
                    write_to_logfile(log, out, err, cmd)
                else:
                    if fasta:
                        ax = 'asm5'
                        # ax = 'sr'
                    # If Nanopore FASTQ:
                    else:
                        ax = 'map-ont'
                    # '> {outsam}'
                    cmd = 'minimap2 --MD -t {threads} -ax {ax} {ref} {reads}' \
                        .format(ax=ax,
                                ref=rmlst_fasta,
                                reads=unpaired_bait,
                                outsam=outsam,
                                threads=threads)
                    # ' | samtools rmdup - -S -' \
                    cmd += ' | samtools view -@ {threads} -h -bT {abs_ref_link} -' \
                           ' | samtools sort - -@ {threads} -o {sorted_bam}' \
                        .format(threads=threads,
                                abs_ref_link=rmlst_fasta,
                                sorted_bam=sorted_bam)
                    out, err = run_cmd(cmd)
                    write_to_logfile(log, out, err, cmd)

        if not os.path.isfile(sorted_bam):
            # noinspection PyUnresolvedReferences
            pysam.sort('-o', sorted_bam, outbam)
        if not os.path.isfile(sorted_bam + '.bai'):
            pysam.index(sorted_bam)
        # Now find number of multi-positions for each rMLST gene/allele combination
        multi_positions = 0

        # Run the BAM parsing in parallel! Some refactoring of the code would likely be a good idea so this
        # isn't quite so ugly, but it works.
        p = multiprocessing.Pool(processes=threads)
        nanopore = True if data_type == 'Nanopore' else False
        nanopore_list = [nanopore] * len(gene_alleles)
        allele_records = SeqIO.to_dict(SeqIO.parse(rmlst_fasta, 'fasta'))
        bamfile_list = [sorted_bam] * len(gene_alleles)
        reference_fasta_list = [rmlst_fasta] * len(gene_alleles)
        fasta_list = [fasta] * len(gene_alleles)
        quality_cutoff_list = [quality_cutoff] * len(gene_alleles)
        base_cutoff_list = [base_cutoff] * len(gene_alleles)
        base_fraction_list = [base_fraction_cutoff] * len(gene_alleles)
        records_list = [allele_records] * len(gene_alleles)
        fastq_records_list = [fastq_records] * len(gene_alleles)
        error_cutoff_list = [error_cutoff] * len(gene_alleles)
        multibase_dict_list = list()
        report_write_list = list()
        if debug == 'debug':
            for i, gene in enumerate(gene_alleles):
                multibase_dict, report_write = read_contig(
                    contig_name=gene,
                    bamfile_name=bamfile_list[i],
                    reference_fasta=reference_fasta_list[i],
                    allele_records=records_list[i],
                    fastq_records=fastq_records_list[i],
                    quality_cutoff=quality_cutoff_list[i],
                    base_cutoff=base_cutoff_list[i],
                    base_fraction_cutoff=base_fraction_list[i],
                    fasta=fasta_list[i],
                    nanopore = nanopore_list[i],
                    error_cutoff=error_cutoff_list[i])
                multibase_dict_list.append(multibase_dict)
                report_write_list.append(report_write)
        else:
            for multibase_dict, report_write in p.starmap(read_contig,
                                                          zip(gene_alleles,
                                                              bamfile_list,
                                                              reference_fasta_list,
                                                              records_list,
                                                              fastq_records_list,
                                                              quality_cutoff_list,
                                                              base_cutoff_list,
                                                              base_fraction_list,
                                                              fasta_list,
                                                              error_cutoff_list,
                                                              nanopore_list),
                                                          chunksize=1):
                multibase_dict_list.append(multibase_dict)
                report_write_list.append(report_write)
            p.close()
            p.join()
    except SamtoolsError:
        pysam_pass = False
        multi_positions = 0
        multibase_dict_list = list()
        report_write_list = list()
    # Write out report info.
    report_file = os.path.join(output_folder, sample_name + '_contamination.csv')
    with open(report_file, 'w') as r:
        r.write('Gene,Position,RefBase,CongruentSNVs,TotalSNVs,ForwardSNVs,ReverseSNVs,SNVCoverage,TotalCoverage,'
                'BaseCutoff,ErrorPercent\n')
        for item in report_write_list:
            for contamination_info in item:
                r.write(contamination_info)
    # Total up the number of multibase positions.
    for multibase_position_dict in multibase_dict_list:
        multi_positions += sum([len(snp_positions) for snp_positions in multibase_position_dict.values()])
    if cgmlst_db is None:
        snp_cutoff = math.ceil(rmlst_gene_length / 10000) + 1
    elif fasta:
        snp_cutoff = 1
    else:
        snp_cutoff = 10

    logging.info('Done! Number of contaminating SNVs found: {}\n'.format(multi_positions))
    write_output(output_report=os.path.join(output_folder, 'confindr_report.csv'),
                 sample_name=sample_name,
                 multi_positions=multi_positions,
                 genus=genus,
                 total_gene_length=rmlst_gene_length,
                 snp_cutoff=snp_cutoff,
                 database_download_date=database_download_date,
                 pysam_pass=pysam_pass)
    if keep_files is False:
        shutil.rmtree(sample_tmp_dir)


def write_output(output_report, sample_name, multi_positions, genus, total_gene_length,
                 database_download_date, snp_cutoff=3, pysam_pass=True):
    """
    Function that writes the output generated by ConFindr to a report file. Appends to a file that already exists,
    or creates the file if it doesn't already exist.
    :param output_report: Path to CSV output report file. Should have headers SampleName,Genus,NumContamSNVs, and
    ContamStatus, in that order.
    :param sample_name: string - name of sample
    :param multi_positions: integer - number of positions that were found to have more than one base present.
    :param genus: string - The genus of your sample
    :param total_gene_length: integer - number of bases examined to make a contamination call.
    :param database_download_date:
    :param snp_cutoff: Number of cSNVs to use to call a sample contaminated. Default 3. (INT)
    :param pysam_pass: Boolean of whether pysam encountered an error
    """
    # If the report file hasn't been created, make it, with appropriate header.
    if not os.path.isfile(output_report):
        with open(os.path.join(output_report), 'w') as f:
            f.write('Sample,Genus,NumContamSNVs,ContamStatus,'
                    'BasesExamined,DatabaseDownloadDate\n')
    if pysam_pass:
        if multi_positions >= snp_cutoff or len(genus.split(':')) > 1:
            contaminated = True
        else:
            contaminated = False
    else:
        contaminated = 'Pysam SamtoolsError'
        multi_positions = 'ND'
    with open(output_report, 'a+') as f:
        f.write('{samplename},{genus},{numcontamsnvs},{contamstatus},'
                '{gene_length},{database_download_date}\n'.format(samplename=sample_name,
                                                                  genus=genus,
                                                                  numcontamsnvs=multi_positions,
                                                                  contamstatus=contaminated,
                                                                  gene_length=total_gene_length,
                                                                  database_download_date=database_download_date))


def check_for_databases_and_download(database_location):
    # Check for the files necessary - should have rMLST_combined.fasta, gene_allele.txt, profiles.txt, and refseq.msh
    necessary_files = ['Escherichia_db_cgderived.fasta', 'Listeria_db_cgderived.fasta',
                       'Salmonella_db_cgderived.fasta', 'refseq.msh']
    optional_files = ['rMLST_combined.fasta', 'gene_allele.txt', 'profiles.txt']
    all_files_present = True
    for necessary_file in necessary_files:
        if not os.path.isfile(os.path.join(database_location, necessary_file)):
            logging.warning('Could not find {}'.format(necessary_file))
            all_files_present = False

    if not all_files_present:
        logging.warning('Databases not present - downloading basic databases now...')
        if not os.path.isdir(database_location):
            os.makedirs(database_location)
        download_mash_sketch(database_location)
        download_cgmlst_derived_data(database_location)

    optional_files_present = True
    for optional_file in optional_files:
        if not os.path.isfile(os.path.join(database_location, optional_file)):
            optional_files_present = False
    if not optional_files_present:
        logging.warning('Did not find rMLST databases, if you want to use ConFindr on genera other than Listeria, '
                        'Salmonella, and Escherichia, you\'ll need to download them. Instructions are available at '
                        'https://olc-bioinformatics.github.io/ConFindr/install/#downloading-confindr-databases\n')


def check_valid_base_fraction(base_fraction):
    """
    Checks that base fraction specified is either None, and therefore won't be used, or is between 0 and 1.
    :param base_fraction: Base fraction, which should be either None or a float.
    :return: True if base fraction is valid, False if not.
    """
    if base_fraction is None:
        return True
    if 0 <= base_fraction <= 1:
        return True
    else:
        return False


def check_acceptable_xmx(xmx_string):
    """
    BBTools can have their memory set manually. This will check that the memory setting is actually valid
    :param xmx_string: The users requested XMX, as a string.
    :return: True if the Xmx string will be accepted by BBTools, otherwise false.
    """
    acceptable_xmx = True
    acceptable_suffixes = ['K', 'M', 'G']
    if xmx_string[-1].upper() not in acceptable_suffixes:
        acceptable_xmx = False
        logging.error('ERROR: Memory must be specified as K (kilobytes), M (megabytes), or G (gigabytes). '
                      'Your specified suffix was {}.'.format(xmx_string[-1]))
    if '.' in xmx_string:
        acceptable_xmx = False
        logging.error('ERROR: Xmx strings must be integers, floating point numbers are not accepted.')
    if not str.isdigit(xmx_string[:-1]):
        acceptable_xmx = False
        logging.error('ERROR: The amount of memory requested was not an integer.')
    return acceptable_xmx


def get_version():
    try:
        version = 'ConFindr {}'.format(pkg_resources.get_distribution('confindr').version)
    except pkg_resources.DistributionNotFound:
        version = 'ConFindr (Unknown version)'
    return version



#!/usr/bin/env python
from confindr_src.methods import download_cgmlst_derived_data, download_mash_sketch, index
from rauth import OAuth1Session
from Bio import SeqIO
import argparse
import datetime
import logging
import shutil
import glob
import ssl
import csv
import re
import os

class RmlstRest(object):

    def get_session_token(self):
        session_request = OAuth1Session(self.consumer_key,
                                        self.consumer_secret,
                                        access_token=self.access_token,
                                        access_token_secret=self.access_secret)
        url = self.test_rest_url + '/oauth/get_session_token'
        if self.unverified:
            r = session_request.get(url, verify=False)
        else:
            r = session_request.get(url)
        if r.status_code == 200:
            self.session_token = r.json()['oauth_token']
            self.session_secret = r.json()['oauth_token_secret']
        else:
            logging.error('ERROR: Couldn\'t get a session token for rMLST database download. Check that your consumer '
                          'secret and access token files have valid credentials and try again.')
            quit(code=1)

    def get_loci_and_scheme_url(self):
        session = OAuth1Session(self.consumer_key,
                                self.consumer_secret,
                                access_token=self.session_token,
                                access_token_secret=self.session_secret)
        if self.unverified:
            r = session.get(self.test_rest_url, verify=False)
        else:
            r = session.get(self.test_rest_url)
        if r.status_code == 200 or r.status_code == 201:
            if re.search('json', r.headers['content-type'], flags=0):
                decoded = r.json()
            else:
                decoded = r.text
            # Extract the URLs from the returned data
            self.loci = decoded['loci']
            self.profile = decoded['schemes']
        else:
            logging.error('ERROR: Could not find URLs for rMLST download, they may have moved. Please open an issue '
                          'at https://github.com/OLC-Bioinformatics/ConFindr/issues and we\'ll get things sorted out.')
            quit(code=1)

    def download_loci(self):
        session = OAuth1Session(self.consumer_key,
                                self.consumer_secret,
                                access_token=self.session_token,
                                access_token_secret=self.session_secret)
        if self.unverified:
            r = session.get(self.loci, verify=False)
        else:
            r = session.get(self.loci)
        if r.status_code == 200 or r.status_code == 201:
            if re.search('json', r.headers['content-type'], flags=0):
                decoded = r.json()
            else:
                decoded = r.text
            # Extract all the URLs in the decoded dictionary under the key 'loci'
            for locus_url in decoded['loci']:
                output_file = os.path.join(self.output_folder, '{}.tfa'.format(os.path.split(locus_url)[1]))
                logging.info('Downloading {}...'.format(os.path.split(locus_url)[1]))
                with open(output_file, 'w') as f:
                    if self.unverified:
                        download = session.get(locus_url + '/alleles_fasta', verify=False)
                    else:
                        download = session.get(locus_url + '/alleles_fasta')
                    if download.status_code == 200 or download.status_code == 201:
                        if re.search('json', download.headers['content-type'], flags=0):
                            decoded = download.json()
                        else:
                            decoded = download.text
                        with open(output_file, 'w') as locus_fasta:
                            locus_fasta.write(decoded)
                
        else:
            logging.error('ERROR: Could not find URLs for rMLST download, they may have moved. Please open an issue '
                          'at https://github.com/OLC-Bioinformatics/ConFindr/issues and we\'ll get things sorted out.')
            quit(code=1)

    def download_profile(self):
        profile_file = os.path.join(self.output_folder, 'profiles.txt')
        session = OAuth1Session(self.consumer_key,
                                self.consumer_secret,
                                access_token=self.session_token,
                                access_token_secret=self.session_secret)
        if self.unverified:
            r = session.get(self.profile + '/1/profiles_csv', verify=False)
        else:
            r = session.get(self.profile + '/1/profiles_csv')
        logging.info('Downloading rMLST profiles...')
        if r.status_code == 200 or r.status_code == 201:
            if re.search('json', r.headers['content-type'], flags=0):
                decoded = r.json()
            else:
                decoded = r.text
            # Write the profile file to disk
            with open(profile_file, 'w') as profile:
                profile.write(decoded)

    def get_request_token(self):
        session = OAuth1Session(consumer_key=self.consumer_key,
                                consumer_secret=self.consumer_secret)
        # Use the test URL in the GET request
        r = session.request(method='GET',
                            url=self.request_token_url,
                            params={'oauth_callback': 'oob'})
        if r.status_code == 200:
            self.request_token = r.json()['oauth_token']
            self.request_secret = r.json()['oauth_token_secret']

    def get_access_token(self):
        authorize_url = self.test_web_url + '&page=authorizeClient&oauth_token=' + self.request_token
        print('Visit this URL in your browser: ' + authorize_url)
        verifier = input('Enter oauth_verifier from browser: ')
        session_request = OAuth1Session(consumer_key=self.consumer_key,
                                        consumer_secret=self.consumer_secret,
                                        access_token=self.request_token,
                                        access_token_secret=self.request_secret)
        # Perform a GET request with the appropriate keys and tokens
        if self.unverified:
            r = session_request.get(self.access_token_url, verify=False,
                                    params={
                                        'oauth_verifier': verifier
                                    })
        else:
            r = session_request.get(self.access_token_url,
                                    params={
                                        'oauth_verifier': verifier
                                    })
        # If the status code is '200' (OK), proceed
        if r.status_code == 200:
            # Save the JSON-decoded token secret and token
            self.access_token = r.json()['oauth_token']
            self.access_secret = r.json()['oauth_token_secret']

    def __init__(self, consumer_secret_file, output_folder, unverified=False):
        self.test_rest_url = 'http://rest.pubmlst.org/db/pubmlst_rmlst_seqdef'
        self.test_web_url = 'http://pubmlst.org/cgi-bin/bigsdb/bigsdb.pl?db=pubmlst_rmlst_seqdef'
        self.request_token_url = self.test_rest_url + '/oauth/get_request_token'
        self.access_token_url = self.test_rest_url + '/oauth/get_access_token'
        self.authorize_url = self.test_web_url + '&page=authorizeClient'
        self.output_folder = output_folder
        self.unverified = unverified
         
        # Get the consumer secret set up.
        if not os.path.isfile(consumer_secret_file):
            logging.error('ERROR: Could not find consumer secret file. Please make sure the file you specified '
                          '({}) exists and try again.'.format(consumer_secret_file))
            quit(code=1)
        with open(consumer_secret_file) as f:
            lines = f.readlines()
        try:
            self.consumer_key = lines[0].rstrip()
            self.consumer_secret = lines[1].rstrip()
        except IndexError:
            logging.error('ERROR: Could not parse your consumer secret file. File should have supplied consumer key '
                          'on first line, and consumer secret on the second line.')
            quit(code=1) 
            
        self.session_secret = str()
        self.session_token = str()
        self.loci = str()
        self.profile = str()
        self.request_token = str()
        self.request_secret = str()
        self.access_token = str()
        self.access_secret = str()


def create_gene_allele_file(profiles_file, gene_allele_file):
    genus_allele_info = dict()
    genera = set()
    with open(profiles_file) as tsvfile:
        reader = csv.DictReader(tsvfile, delimiter='\t')
        for row in reader:
            genus = row['genus']
            # If the genus is uncertain e.g. Escherichia/Shigella, split on the /, and use Escherichia as the genus
            if '/' in genus:
                genus = genus.split('/')[0]
            genera.add(genus)
            if genus not in genus_allele_info:
                genus_allele_info[genus] = list()
            for i in range(1, 66):
                if i < 10:
                    gene = 'BACT00000' + str(i)
                else:
                    gene = 'BACT0000' + str(i)
                if gene in row:
                    allele_number = row[gene]
                    gene_allele = '{}_{}'.format(gene, allele_number)
                    if allele_number != 'N' and gene_allele not in genus_allele_info[genus]:
                        genus_allele_info[genus].append(gene_allele)
    with open(gene_allele_file, 'w') as f:
        for genus in genus_allele_info:
            f.write(str(genus) + ':')
            for allele in genus_allele_info[genus]:
                f.write(str(allele) + ',')
            f.write('\n')
    return genera


def setup_confindr_database(output_folder, consumer_secret, index_databases=False, unverified=False):
    # Go through the REST API in order to get profiles downloaded.
    rmlst_rest = RmlstRest(consumer_secret_file=consumer_secret,
                           output_folder=output_folder, unverified=unverified)
    rmlst_rest.get_request_token()
    rmlst_rest.get_access_token()
    rmlst_rest.get_session_token()
    rmlst_rest.get_loci_and_scheme_url()
    rmlst_rest.download_loci()
    rmlst_rest.download_profile()

    # With the sequences downloaded, make a file of all rMLST sequences combined.
    logging.info('Combining rMLST files...')
    with open(os.path.join(output_folder, 'rMLST_combined.fasta'), 'w') as f:
        locus_files = sorted(glob.glob(os.path.join(output_folder, 'BACT*.tfa')))
        for locus_file in locus_files:
            for record in SeqIO.parse(locus_file, 'fasta'):
                record.id = record.id.replace('-', '_')
                try:
                    record.seq._data = record.seq._data.replace('-', '').replace('N', '')
                except TypeError:
                    record.seq._data = record.seq._data.replace(b'-', b'').replace(b'N', b'')

                # If the entire FASTA sequence is encoded in byte-like
                # formatting (b' at the beginning and ' at the end of the 
                # sequence), fix:
                if record.seq._data[0:2] == "b'" and record.seq._data[-1] == "'":
                    record.seq._data = record.seq._data.replace("b'", "").replace("'", "")
                
                record.name = ''
                record.description = ''
                SeqIO.write(record, f, 'fasta')
            # Clean up individual file.
            try:
                os.remove(locus_file)
            except OSError:
                logging.warning('WARNING: Could not delete {}. This won\'t affect ConFindr performance, but '
                                ' you may want to delete it to save on disk space.'.format(locus_file))

    logging.info('Assigning alleles to genera...')
    # Parse profiles so that we know what alleles are found with each genus.
    genera = create_gene_allele_file(profiles_file=os.path.join(output_folder, 'profiles.txt'),
                                     gene_allele_file=os.path.join(output_folder, 'gene_allele.txt'))
    if index_databases:
        index(output_folder=output_folder,
              genera=sorted(list(genera)),
              cgderived=False)


def main():
    logging.basicConfig(format='\033[92m \033[1m %(asctime)s \033[0m %(message)s ',
                        level=logging.INFO,
                        datefmt='%Y-%m-%d %H:%M:%S')
    parser = argparse.ArgumentParser()
    parser.add_argument('-o', '--output_folder',
                        default=os.environ.get('CONFINDR_DB', os.path.expanduser('~/.confindr_db')),
                        help='Path to download databases to - if folder does not exist, will be created. If folder does'
                             ' exist, will be deleted and updated sequences downloaded. Defaults to ~/.confindr_db, or '
                             'the CONFINDR_DB environmental variable.')
    parser.add_argument('-s', '--secret_file',
                        type=str,
                        help='Path to consumer secret file for rMLST database.')
    parser.add_argument('-i', '--index_databases',
                        action='store_true',
                        help='Enable this option if you are installing the databases to a drive that will be read-only '
                             'after the installation. The script will create and index all the necessary genus-specific'
                             ' database files. Note that this is very slow for the rMLST database.')
    parser.add_argument('-u', '--unverified',
                        action='store_true',
                        help="Enable this option if you plan on running ConFindr behind a firewall and/or have a self- "
                        "signed certificate. Adds 'verify=False' during session requests.")
    args = parser.parse_args()
    if os.path.isdir(args.output_folder):
        logging.info('Removing old databases...')
        shutil.rmtree(args.output_folder)
    os.makedirs(args.output_folder)
    if args.unverified:
        ssl._create_default_https_context = ssl._create_unverified_context
    download_cgmlst_derived_data(args.output_folder)
    if args.secret_file is None:
        logging.warning('WARNING: Without an rMLST secret file, data will only be downloaded for Escherichia, '
                        'Salmonella, and Listeria. See '
                        'https://olc-bioinformatics.github.io/ConFindr/install/#downloading-confindr-databases for '
                        'instructions on how to get access to rMLST databases so ConFindr can be used for other species'
                        ' as well')
    else:
        setup_confindr_database(output_folder=args.output_folder,
                                consumer_secret=args.secret_file,
                                index_databases=args.index_databases,
                                unverified=args.unverified)
    download_mash_sketch(args.output_folder)
    current_year = datetime.datetime.utcnow().year
    current_month = datetime.datetime.utcnow().month
    current_day = datetime.datetime.utcnow().day
    with open(os.path.join(args.output_folder, 'download_date.txt'), 'w') as f:
        f.write('{}-{}-{}'.format(current_year, current_month, current_day))
    logging.info('Done downloading ConFindr databases!')
    
    
if __name__ == '__main__':
    main()



import os
import subprocess
from subprocess import Popen, PIPE


def run_subprocess(command):
    """
    command is the command to run, as a string.
    runs a subprocess, returns stdout and stderr from the subprocess as strings.
    """
    x = Popen(command, shell=True, stdout=PIPE, stderr=PIPE)
    out, err = x.communicate()
    out = out.decode('utf-8')
    err = err.decode('utf-8')
    if x.returncode != 0:
        raise subprocess.CalledProcessError(x.returncode, cmd=command)
    return out, err


def kwargs_to_string(kwargs):
    """
    Given a set of kwargs, turns them into a string which can then be passed to a command.
    :param kwargs: kwargs from a function call.
    :return: outstr: A string, which is '' if no kwargs were given, and the kwargs in string format otherwise.
    """
    outstr = ''
    for arg in kwargs:
        outstr += ' {}={}'.format(arg, kwargs[arg])
    return outstr


def bbmap(reference, forward_in, out_bam, reverse_in='NA', returncmd=False, **kwargs):
    """
    Wrapper for bbmap. Assumes that bbmap executable is in your $PATH.
    :param reference: Reference fasta. Won't be written to disk by default. If you want it to be, add nodisk='t' as an arg.
    :param forward_in: Input reads. Should be in fastq format.
    :param out_bam: Output file. Should end in .sam or .bam
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param reverse_in: If your reverse reads are present and normal conventions (_R1 for forward, _R2 for reverse) are
     followed, the reverse reads will be followed automatically. If you want to specify reverse reads, you may do so.
    :param kwargs: Other arguments to give to bbmap in parameter=argument format. See bbmap documentation for full list.
    :return: out and err: stdout string and stderr string from running bbmap.
    """
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        cmd = 'bbmap.sh ref={} in={} in2={} out={} nodisk{}'.format(reference, forward_in, reverse_in, out_bam, options)
    elif reverse_in == 'NA':
        cmd = 'bbmap.sh ref={} in={} out={} nodisk{}'.format(reference, forward_in, out_bam, options)
    else:
        cmd = 'bbmap.sh ref={} in={} in2={} out={} nodisk{}'.format(reference, forward_in, reverse_in, out_bam, options)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def bbduk_trim(forward_in, forward_out, reverse_in='NA', reverse_out='NA', returncmd=False, **kwargs):
    """
    Wrapper for using bbduk to quality trim reads. Contains arguments used in OLC Assembly Pipeline, but these can
    be overwritten by using keyword parameters.
    :param forward_in: Forward reads you want to quality trim.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param forward_out: Output forward reads.
    :param reverse_in: Reverse input reads. Don't need to be specified if _R1/_R2 naming convention is used.
    :param reverse_out: Reverse output reads. Don't need to be specified if _R1/_R2 convention is used.
    :param kwargs: Other arguments to give to bbduk in parameter=argument format. See bbduk documentation for full list.
    :return: out and err: stdout string and stderr string from running bbduk.
    """
    options = kwargs_to_string(kwargs)
    cmd = 'which bbduk.sh'
    try:
        subprocess.check_output(cmd.split()).decode('utf-8')
    except subprocess.CalledProcessError:
        print('ERROR: Could not find bbduk. Plase check that the bbtools package is installed and on your $PATH.\n\n')
        raise FileNotFoundError
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        if reverse_out == 'NA':
            if '_R1' in forward_out:
                reverse_out = forward_out.replace('_R1', '_R2')
            else:
                raise ValueError('If you do not specify reverse_out, forward_out must contain R1.\n\n')
        cmd = 'bbduk.sh in1={f_in} in2={r_in} out1={f_out} out2={r_out} qtrim=w trimq=20 k=25 minlength=50 ' \
              'forcetrimleft=15 ref=adapters overwrite hdist=1 qin=33 tpe tbo{optn}'\
            .format(f_in=forward_in,
                    r_in=reverse_in,
                    f_out=forward_out,
                    r_out=reverse_out,
                    optn=options)
    elif reverse_in == 'NA':
        cmd = 'bbduk.sh in={f_in} out={f_out} qtrim=w trimq=20 k=25 minlength=50 forcetrimleft=15' \
              ' ref=adapters overwrite hdist=1 qin=33 tpe tbo{optn}'\
            .format(f_in=forward_in,
                    f_out=forward_out,
                    optn=options)
    else:
        if reverse_out == 'NA':
            raise ValueError('Reverse output reads must be specified.')
        cmd = 'bbduk.sh in1={f_in} in2={r_in} out1={f_out} out2={r_out} qtrim=w trimq=20 k=25 minlength=50 ' \
              'forcetrimleft=15 ref=adapters overwrite hdist=1 qin=33 tpe tbo{optn}'\
            .format(f_in=forward_in,
                    r_in=reverse_in,
                    f_out=forward_out,
                    r_out=reverse_out,
                    optn=options)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def tadpole(forward_in, forward_out, reverse_in='NA', returncmd=False, reverse_out='NA', mode='correct', **kwargs):
    """
    Runs tadpole. Default is to run in correction mode, but other modes ('contig', 'extend') can also be specified.
    :param forward_in: Forward input reads.
    :param forward_out: Forward output reads.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param reverse_in: Reverse reads. Only specify if not following _R1/_R2 convention/not in same folder as input.
    :param reverse_out: Reverse output reads. Automatically generated unless specified.
    :param mode: Mode to run tadpole in. Default is 'correct'.
    :param kwargs: Other arguments to give to tadpole in parameter='argument' format. See tadpole documentation for full list.
    :return: out and err: stdout string and stderr string from running tadpole.
    """
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        if reverse_out == 'NA':
            if '_R1' in forward_out:
                reverse_out = forward_out.replace('_R1', '_R2')
            else:
                raise ValueError('If you do not specify reverse_out, forward_out must contain _R1.\n\n')
        cmd = 'tadpole.sh in1={} in2={} out1={} out2={} mode={} {}'.format(forward_in, reverse_in,
                                                                           forward_out, reverse_out,
                                                                           mode, options)
    elif reverse_in == 'NA':
        cmd = 'tadpole.sh in={} out={} mode={} {}'.format(forward_in, forward_out, mode, options)
    else:
        if reverse_out == 'NA':
            raise ValueError('Reverse output reads must be specified.')
        cmd = 'tadpole.sh in1={} in2={} out1={} out2={} mode={} {}'.format(forward_in, reverse_in,
                                                                           forward_out, reverse_out,
                                                                           mode, options)
    if not os.path.isfile(forward_out):
        out, err = run_subprocess(cmd)
    else:
        out = str()
        err = str()
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def bbnorm(forward_in, forward_out, returncmd=False, reverse_in='NA', reverse_out='NA', **kwargs):
    """
    Runs bbnorm to normalize read depth. Default target kmer depth is left at bbnorm's default, which is 100.
    :param forward_in: Forward input reads.
    :param forward_out: Forward output reads.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param reverse_in: Reverse reads. Only specify if not following _R1/_R2 convention/not in same folder as input.
    :param reverse_out: Reverse output reads. Automatically generated unless specified.
    :param kwargs: Other arguments to give to bbnorm in parameter='argument' format. See bbnorm documentation for full list.
    :return: out and err: stdout string and stderr string from running bbnorm.
    """
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        if reverse_out == 'NA':
            if '_R1' in forward_out:
                reverse_out = forward_out.replace('_R1', '_R2')
            else:
                raise ValueError('If you do not specify reverse_out, forward_out must contain _R1.\n\n')
        cmd = 'bbnorm.sh in1={} in2={} out={} out2={} {}'.format(forward_in, reverse_in,
                                                                  forward_out, reverse_out,
                                                                  options)
    elif reverse_in == 'NA':
        cmd = 'bbnorm.sh in={} out={} {}'.format(forward_in, forward_out, options)
    else:
        if reverse_out == 'NA':
            raise ValueError('Reverse output reads must be specified.')
        cmd = 'bbnorm.sh in1={} in2={} out1={} out2={} {}'.format(forward_in, reverse_in,
                                                                  forward_out, reverse_out,
                                                                  options)
    if not os.path.isfile(forward_out):
        out, err = run_subprocess(cmd)
    else:
        out = str()
        err = str()
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def bbmerge(forward_in, merged_reads, returncmd=False, reverse_in='NA', **kwargs):
    """
    Runs bbmerge.
    :param forward_in: Forward input reads. Reverse reads automatically detected if present in the same folder.
    :param merged_reads: Output file to write merged reads to.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param reverse_in: Reverse input file, if you don't want it autodetected.
    :param kwargs: Other arguments to give to bbmerge in parameter='argument' format. See bbmerge documentation for full list.
    :return: out and err: stdout string and stderr string from running bbmerge.
    """
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        cmd = 'bbmerge.sh in={} in2={} out={} {}'.format(forward_in, reverse_in, merged_reads, options)
    elif reverse_in == 'NA':
        cmd = 'bbmerge.sh in={} out={} {}'.format(forward_in, merged_reads, options)
    else:
        cmd = 'bbmerge.sh in={} in2={} out={} {}'.format(forward_in, reverse_in, merged_reads, options)
    if not os.path.isfile(merged_reads):
        out, err = run_subprocess(cmd)
    else:
        out = str()
        err = str()
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def bbduk_bait(reference, forward_in, forward_out, returncmd=False, reverse_in='NA', reverse_out='NA', **kwargs):
    """
    Uses bbduk to bait out reads that have kmers matching to a reference.
    :param reference: Reference you want to pull reads out for. Should be in fasta format.
    :param forward_in: Forward reads you want to quality trim.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param forward_out: Output forward reads.
    :param reverse_in: Reverse input reads. Don't need to be specified if _R1/_R2 naming convention is used.
    :param reverse_out: Reverse output reads. Don't need to be specified if _R1/_R2 convention is used.
    :param kwargs: Other arguments to give to bbduk in parameter=argument format. See bbduk documentation for full list.
    :return: out and err: stdout string and stderr string from running bbduk.
    """
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        if reverse_out == 'NA':
            if '_R1' in forward_out:
                reverse_out = forward_out.replace('_R1', '_R2')
            else:
                raise ValueError('If you do not specify reverse_out, forward_out must contain _R1.\n\n')
        cmd = 'bbduk.sh qin=33 in={} in2={} outm={} outm2={} ref={}{}'.format(forward_in, reverse_in,
                                                                       forward_out, reverse_out,
                                                                       reference, options)
    elif reverse_in == 'NA':
        cmd = 'bbduk.sh qin=33 in={} outm={} ref={}{}'.format(forward_in, forward_out, reference, options)
    else:
        if reverse_out == 'NA':
            raise ValueError('Reverse output reads must be specified.')
        cmd = 'bbduk.sh qin=33 in={} in2={} outm={} outm2={} ref={}{}'.format(forward_in, reverse_in,
                                                                       forward_out, reverse_out,
                                                                       reference, options)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def bbduk_filter(reference, forward_in, forward_out, returncmd=False, reverse_in='NA', reverse_out='NA', **kwargs):
    """
    Uses bbduk to filter out reads that have kmers matching to a reference.
    :param reference: Reference you want to pull reads out for. Should be in fasta format.
    :param forward_in: Forward reads you want to quality trim.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param forward_out: Output forward reads.
    :param reverse_in: Reverse input reads. Don't need to be specified if _R1/_R2 naming convention is used.
    :param reverse_out: Reverse output reads. Don't need to be specified if _R1/_R2 convention is used.
    :param kwargs: Other arguments to give to bbduk in parameter=argument format. See bbduk documentation for full list.
    :return: out and err: stdout string and stderr string from running bbduk.
    """
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        if reverse_out == 'NA':
            if '_R1' in forward_out:
                reverse_out = forward_out.replace('_R1', '_R2')
            else:
                raise ValueError('If you do not specify reverse_out, forward_out must contain _R1.\n\n')
        cmd = 'bbduk.sh qin=33 in={} in2={} out={} out2={} ref={}{}'.format(forward_in, reverse_in,
                                                                     forward_out, reverse_out,
                                                                     reference, options)
    elif reverse_in == 'NA':
        cmd = 'bbduk.sh qin=33 in={} out={} ref={}{}'.format(forward_in, forward_out, reference, options)
    else:
        if reverse_out == 'NA':
            raise ValueError('Reverse output reads must be specified.')
        cmd = 'bbduk.sh qin=33 in={} in2={} out={} out2={} ref={}{}'.format(forward_in, reverse_in,
                                                                     forward_out, reverse_out,
                                                                     reference, options)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def dedupe(input_file, output_file, returncmd=False, **kwargs):
    """
    Runs dedupe from the bbtools package.
    :param input_file: Input file.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param output_file: Output file.
    :param kwargs: Arguments to give to dedupe in parameter=argument format. See dedupe documentation for full list.
    :return: out and err: stdout string and stderr string from running dedupe.
    """
    options = kwargs_to_string(kwargs)
    cmd = 'dedupe.sh in={} out={}{}'.format(input_file, output_file, options)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def seal(reference, forward_in, output_file, reverse_in='NA', returncmd=False, **kwargs):
    """
    Runs seal from the bbtools package.
    :param reference: Reference file, in fasta format.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param forward_in: Forward reads, fastq format.
    :param output_file: Output file to put rpkm statistics into.
    :param reverse_in: Reverse reads. Not necessary to specify if in same folder and follow _R1/_R2 convention.
    :param kwargs: Arguments to give to seal in parameter=argument format. See seal documentation for full list.
    :return: out and err: stdout string and stderr string from running seal.
    """
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        cmd = 'seal.sh ref={} in={} in2={} rpkm={} nodisk{}'.format(reference, forward_in, reverse_in, output_file, options)
    elif reverse_in == 'NA':
        cmd = 'seal.sh ref={} in={} rpkm={} nodisk{}'.format(reference, forward_in, output_file, options)
    else:
        cmd = 'seal.sh ref={} in={} in2={} rpkm={} nodisk{}'.format(reference, forward_in, reverse_in, output_file, options)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def kmercountexact(forward_in, reverse_in='NA', returncmd=False, **kwargs):
    """
    Wrapper for kmer count exact.
    :param forward_in: Forward input reads.
    :param reverse_in: Reverse input reads. Found automatically for certain conventions.
    :param returncmd: If set to true, function will return the cmd string passed to subprocess as a third value.
    :param kwargs: Arguments to give to kmercountexact in parameter='argument' format.
    See kmercountexact documentation for full list.
    :return: out and err: stdout string and stderr string from running kmercountexact.
    """
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        cmd = 'kmercountexact.sh in={} in2={} {}'.format(forward_in, reverse_in, options)
    elif reverse_in == 'NA':
        cmd = 'kmercountexact.sh in={} {}'.format(forward_in, options)
    else:
        cmd = 'kmercountexact.sh in={} in2={} {}'.format(forward_in, reverse_in, options)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def genome_size(peaks_file, haploid=True):
    """
    Finds the genome size of an organsim, based on the peaks file created by kmercountexact.sh
    :param peaks_file: Path to peaks file created by kmercountexact.
    :param haploid: Set to True if organism of interest is haploid, False if not. Default True.
    :return: size of genome, as an int. If size could not be found, return will be 0.
    """
    size = 0
    with open(peaks_file) as peaks:
        lines = peaks.readlines()
    for line in lines:
        if haploid:
            if '#haploid_genome_size' in line:
                size = int(line.split()[1])
        else:
            if '#genome_size' in line:
                size = int(line.split()[1])
    return size


def subsample_reads(forward_in, forward_out, num_bases, returncmd=False, reverse_in='NA', reverse_out='NA',
                    **kwargs):
    options = kwargs_to_string(kwargs)
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        if reverse_out == 'NA':
            if '_R1' in forward_out:
                reverse_out = forward_out.replace('_R1', '_R2')
            else:
                raise ValueError('If you do not specify reverse_out, forward_out must contain _R1.\n\n')
        cmd = 'reformat.sh in1={} in2={} out1={} out2={} samplebasestarget={} {}'.format(forward_in, reverse_in,
                                                                                         forward_out, reverse_out,
                                                                                         str(num_bases), options)
    elif reverse_in == 'NA':
        cmd = 'reformat.sh in={} out={} samplebasestarget={} {}'.format(forward_in, forward_out,
                                                                       str(num_bases), options)
    else:
        if reverse_out == 'NA':
            raise ValueError('Reverse output reads must be specified.')
        cmd = 'reformat.sh in1={} in2={} out1={} out2={} samplebasestarget={} {}'.format(forward_in, reverse_in,
                                                                                         forward_out, reverse_out,
                                                                                         str(num_bases), options)
    if not os.path.isfile(forward_out):
        out, err = run_subprocess(cmd)
    else:
        out = str()
        err = str()
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def validate_reads(forward_in, returncmd=False, reverse_in='NA'):
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        cmd = 'reformat.sh in1={} in2={} vpair'.format(forward_in, reverse_in)
    elif reverse_in == 'NA':
        cmd = 'reformat.sh in={}'.format(forward_in)
    out, err = run_subprocess(cmd)
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def reformat_reads(forward_in, forward_out, returncmd=False, reverse_in='NA', reverse_out='NA'):
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        if reverse_out == 'NA':
            if '_R1' in forward_out:
                reverse_out = forward_out.replace('_R1', '_R2')
            else:
                raise ValueError('If you do not specify reverse_out, forward_out must contain _R1.\n\n')
        cmd = 'reformat.sh in1={} in2={} out1={} out2={} tossbrokenreads=t ow=t'\
            .format(forward_in, reverse_in, forward_out, reverse_out)
    elif reverse_in == 'NA':
        cmd = 'reformat.sh in={} out={} tossbrokenreads=t ow=t'.format(forward_in, forward_out)
    else:
        if reverse_out == 'NA':
            raise ValueError('Reverse output reads must be specified.')
        cmd = 'reformat.sh in1={} in2={} out1={} out2={} tossbrokenreads=t ow=t'\
            .format(forward_in, reverse_in, forward_out, reverse_out)
    if not os.path.isfile(forward_out):
        out, err = run_subprocess(cmd)
    else:
        out = str()
        err = str()
    if returncmd:
        return out, err, cmd
    else:
        return out, err


def repair_reads(forward_in, forward_out, returncmd=False, reverse_in='NA', reverse_out='NA'):
    if os.path.isfile(forward_in.replace('_R1', '_R2')) and reverse_in == 'NA' and '_R1' in forward_in:
        reverse_in = forward_in.replace('_R1', '_R2')
        if reverse_out == 'NA':
            if '_R1' in forward_out:
                reverse_out = forward_out.replace('_R1', '_R2')
            else:
                raise ValueError('If you do not specify reverse_out, forward_out must contain _R1.\n\n')
        cmd = 'repair.sh in1={} in2={} out1={} out2={} tossbrokenreads=t repair=t overwrite=t'\
            .format(forward_in, reverse_in, forward_out, reverse_out)
    else:
        if reverse_out == 'NA':
            raise ValueError('Reverse output reads must be specified.')
        cmd = 'repair.sh in1={} in2={} out1={} out2={} tossbrokenreads=t repair=t overwrite=t'\
            .format(forward_in, reverse_in, forward_out, reverse_out)
    if not os.path.isfile(forward_out):
        out, err = run_subprocess(cmd)
    else:
        out = str()
        err = str()
    if returncmd:
        return out, err, cmd
    else:
        return out, err


